from collections.abc import Sequence
from typing import Optional, Union

import dagster._check as check
import graphene
from dagster._core.definitions.events import AssetKey, AssetPartitionWipeRange
from dagster._core.definitions.partition_key_range import PartitionKeyRange
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.nux import get_has_seen_nux, set_nux_seen
from dagster._core.workspace.permissions import Permissions
from dagster._daemon.asset_daemon import set_auto_materialize_paused

from dagster_graphql.implementation.execution import (
    delete_pipeline_run,
    report_runless_asset_events,
    terminate_pipeline_execution,
    terminate_pipeline_execution_for_runs,
    wipe_assets,
)
from dagster_graphql.implementation.execution.backfill import (
    cancel_partition_backfill,
    create_and_launch_partition_backfill,
    resume_partition_backfill,
    retry_partition_backfill,
)
from dagster_graphql.implementation.execution.dynamic_partitions import (
    add_dynamic_partition,
    delete_dynamic_partitions,
)
from dagster_graphql.implementation.execution.launch_execution import (
    launch_pipeline_execution,
    launch_pipeline_reexecution,
    launch_reexecution_from_parent_run,
)
from dagster_graphql.implementation.external import fetch_workspace, get_full_remote_job_or_raise
from dagster_graphql.implementation.telemetry import log_ui_telemetry_event
from dagster_graphql.implementation.utils import (
    ExecutionMetadata,
    ExecutionParams,
    UserFacingGraphQLError,
    assert_permission_for_asset_graph,
    assert_permission_for_location,
    capture_error,
    check_permission,
    pipeline_selector_from_graphql,
    require_permission_check,
)
from dagster_graphql.schema.backfill import (
    GrapheneAssetPartitionRange,
    GrapheneCancelBackfillResult,
    GrapheneLaunchBackfillResult,
    GrapheneResumeBackfillResult,
)
from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.errors import (
    GrapheneAssetNotFoundError,
    GrapheneConflictingExecutionParamsError,
    GrapheneError,
    GraphenePresetNotFoundError,
    GraphenePythonError,
    GrapheneReloadNotSupported,
    GrapheneRepositoryLocationNotFound,
    GrapheneRunNotFoundError,
    GrapheneUnauthorizedError,
    GrapheneUnsupportedOperationError,
)
from dagster_graphql.schema.external import GrapheneWorkspace, GrapheneWorkspaceLocationEntry
from dagster_graphql.schema.inputs import (
    GrapheneExecutionParams,
    GrapheneLaunchBackfillParams,
    GraphenePartitionsByAssetSelector,
    GrapheneReexecutionParams,
    GrapheneReportRunlessAssetEventsParams,
    GrapheneRepositorySelector,
)
from dagster_graphql.schema.partition_sets import (
    GrapheneAddDynamicPartitionResult,
    GrapheneDeleteDynamicPartitionsResult,
)
from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
from dagster_graphql.schema.runs import (
    GrapheneLaunchMultipleRunsResult,
    GrapheneLaunchMultipleRunsResultOrError,
    GrapheneLaunchRunReexecutionResult,
    GrapheneLaunchRunResult,
    GrapheneLaunchRunSuccess,
    parse_run_config_input,
)
from dagster_graphql.schema.schedule_dry_run import GrapheneScheduleDryRunMutation
from dagster_graphql.schema.schedules import (
    GrapheneResetScheduleMutation,
    GrapheneStartScheduleMutation,
    GrapheneStopRunningScheduleMutation,
)
from dagster_graphql.schema.sensor_dry_run import GrapheneSensorDryRunMutation
from dagster_graphql.schema.sensors import (
    GrapheneResetSensorMutation,
    GrapheneSetSensorCursorMutation,
    GrapheneStartSensorMutation,
    GrapheneStopSensorMutation,
)
from dagster_graphql.schema.util import ResolveInfo, non_null_list


def create_execution_params(graphene_info, graphql_execution_params):
    preset_name = graphql_execution_params.get("preset")
    selector = pipeline_selector_from_graphql(graphql_execution_params["selector"])
    if preset_name:
        if graphql_execution_params.get("runConfigData"):
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(conflicting_param="runConfigData")
            )

        if graphql_execution_params.get("mode"):
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(conflicting_param="mode")
            )

        if selector.op_selection:
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(
                    conflicting_param="selector.solid_selection"
                )
            )

        external_pipeline = get_full_remote_job_or_raise(
            graphene_info,
            selector,
        )

        if not external_pipeline.has_preset(preset_name):
            raise UserFacingGraphQLError(
                GraphenePresetNotFoundError(preset=preset_name, selector=selector)
            )

        preset = external_pipeline.get_preset(preset_name)

        return ExecutionParams(
            selector=selector.with_op_selection(preset.op_selection),
            run_config=preset.run_config,
            mode=preset.mode,
            execution_metadata=create_execution_metadata(
                graphql_execution_params.get("executionMetadata")
            ),
            step_keys=graphql_execution_params.get("stepKeys"),
        )

    return execution_params_from_graphql(graphql_execution_params)


def execution_params_from_graphql(graphql_execution_params):
    return ExecutionParams(
        selector=pipeline_selector_from_graphql(graphql_execution_params.get("selector")),
        run_config=parse_run_config_input(  # pyright: ignore[reportArgumentType]
            graphql_execution_params.get("runConfigData") or {},
            raise_on_error=True,
        ),
        mode=graphql_execution_params.get("mode"),
        execution_metadata=create_execution_metadata(
            graphql_execution_params.get("executionMetadata")
        ),
        step_keys=graphql_execution_params.get("stepKeys"),
    )


def create_execution_metadata(graphql_execution_metadata):
    return (
        ExecutionMetadata(
            run_id=graphql_execution_metadata.get("runId"),
            tags={t["key"]: t["value"] for t in graphql_execution_metadata.get("tags", [])},
            root_run_id=graphql_execution_metadata.get("rootRunId"),
            parent_run_id=graphql_execution_metadata.get("parentRunId"),
        )
        if graphql_execution_metadata
        else ExecutionMetadata(run_id=None, tags={})
    )


class GrapheneDeletePipelineRunSuccess(graphene.ObjectType):
    """Output indicating that a run was deleted."""

    runId = graphene.NonNull(graphene.String)

    class Meta:
        name = "DeletePipelineRunSuccess"


class GrapheneDeletePipelineRunResult(graphene.Union):
    """The output from deleting a run."""

    class Meta:
        types = (
            GrapheneDeletePipelineRunSuccess,
            GrapheneUnauthorizedError,
            GraphenePythonError,
            GrapheneRunNotFoundError,
        )
        name = "DeletePipelineRunResult"


class GrapheneDeleteRunMutation(graphene.Mutation):
    """Deletes a run from storage."""

    Output = graphene.NonNull(GrapheneDeletePipelineRunResult)

    class Arguments:
        runId = graphene.NonNull(graphene.String)

    class Meta:
        name = "DeleteRunMutation"

    @capture_error
    @require_permission_check(Permissions.DELETE_PIPELINE_RUN)
    def mutate(
        self, graphene_info: ResolveInfo, runId: str
    ) -> Union[GrapheneRunNotFoundError, GrapheneDeletePipelineRunSuccess]:
        return delete_pipeline_run(graphene_info, runId)


class GrapheneTerminatePipelineExecutionSuccess(graphene.Interface):
    """Interface indicating that a run was terminated."""

    run = graphene.Field(graphene.NonNull(GrapheneRun))

    class Meta:
        name = "TerminatePipelineExecutionSuccess"


class GrapheneTerminateRunSuccess(graphene.ObjectType):
    """Output indicating that a run was terminated."""

    run = graphene.Field(graphene.NonNull(GrapheneRun))

    class Meta:
        interfaces = (GrapheneTerminatePipelineExecutionSuccess,)
        name = "TerminateRunSuccess"


class GrapheneTerminatePipelineExecutionFailure(graphene.Interface):
    """Interface indicating that a run failed to terminate."""

    run = graphene.NonNull(GrapheneRun)
    message = graphene.NonNull(graphene.String)

    class Meta:
        name = "TerminatePipelineExecutionFailure"


class GrapheneTerminateRunFailure(graphene.ObjectType):
    """Output indicating that a run failed to terminate."""

    run = graphene.NonNull(GrapheneRun)
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneTerminatePipelineExecutionFailure,)
        name = "TerminateRunFailure"


class GrapheneTerminateRunResult(graphene.Union):
    """The output from a run termination."""

    class Meta:
        types = (
            GrapheneTerminateRunSuccess,
            GrapheneTerminateRunFailure,
            GrapheneRunNotFoundError,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "TerminateRunResult"


class GrapheneTerminateRunsResult(graphene.ObjectType):
    """Indicates the runs that successfully terminated and those that failed to terminate."""

    terminateRunResults = non_null_list(GrapheneTerminateRunResult)

    class Meta:
        name = "TerminateRunsResult"


class GrapheneTerminateRunsResultOrError(graphene.Union):
    """The output from runs termination."""

    class Meta:
        types = (GrapheneTerminateRunsResult, GraphenePythonError)
        name = "TerminateRunsResultOrError"


def create_execution_params_and_launch_pipeline_exec(graphene_info, execution_params_dict):
    execution_params = create_execution_params(graphene_info, execution_params_dict)
    assert_permission_for_location(
        graphene_info,
        Permissions.LAUNCH_PIPELINE_EXECUTION,
        execution_params.selector.location_name,
    )
    return launch_pipeline_execution(
        graphene_info,
        execution_params,
    )


class GrapheneLaunchRunMutation(graphene.Mutation):
    """Launches a job run."""

    Output = graphene.NonNull(GrapheneLaunchRunResult)

    class Arguments:
        executionParams = graphene.NonNull(GrapheneExecutionParams)

    class Meta:
        name = "LaunchRunMutation"

    @capture_error
    @require_permission_check(Permissions.LAUNCH_PIPELINE_EXECUTION)
    def mutate(
        self, graphene_info: ResolveInfo, executionParams: GrapheneExecutionParams
    ) -> Union[GrapheneLaunchRunSuccess, GrapheneError, GraphenePythonError]:
        return create_execution_params_and_launch_pipeline_exec(graphene_info, executionParams)


class GrapheneLaunchMultipleRunsMutation(graphene.Mutation):
    """Launches multiple job runs."""

    Output = graphene.NonNull(GrapheneLaunchMultipleRunsResultOrError)

    class Arguments:
        executionParamsList = non_null_list(GrapheneExecutionParams)

    class Meta:
        name = "LaunchMultipleRunsMutation"

    @capture_error
    def mutate(
        self, graphene_info: ResolveInfo, executionParamsList: list[GrapheneExecutionParams]
    ) -> Union[
        GrapheneLaunchMultipleRunsResult,
        GrapheneError,
        GraphenePythonError,
    ]:
        launch_multiple_runs_result = []

        for execution_params in executionParamsList:
            result = GrapheneLaunchRunMutation.mutate(
                None, graphene_info, executionParams=execution_params
            )
            launch_multiple_runs_result.append(result)

        return GrapheneLaunchMultipleRunsResult(
            launchMultipleRunsResult=launch_multiple_runs_result
        )


class GrapheneLaunchBackfillMutation(graphene.Mutation):
    """Launches a set of partition backfill runs."""

    Output = graphene.NonNull(GrapheneLaunchBackfillResult)

    class Arguments:
        backfillParams = graphene.NonNull(GrapheneLaunchBackfillParams)

    class Meta:
        name = "LaunchBackfillMutation"

    @capture_error
    @require_permission_check(Permissions.LAUNCH_PARTITION_BACKFILL)
    def mutate(self, graphene_info: ResolveInfo, backfillParams: GrapheneLaunchBackfillParams):
        return create_and_launch_partition_backfill(graphene_info, backfillParams)


class GrapheneCancelBackfillMutation(graphene.Mutation):
    """Cancels a set of partition backfill runs."""

    Output = graphene.NonNull(GrapheneCancelBackfillResult)

    class Arguments:
        backfillId = graphene.NonNull(graphene.String)

    class Meta:
        name = "CancelBackfillMutation"

    @capture_error
    @require_permission_check(Permissions.CANCEL_PARTITION_BACKFILL)
    def mutate(self, graphene_info: ResolveInfo, backfillId: str):
        return cancel_partition_backfill(graphene_info, backfillId)


class GrapheneResumeBackfillMutation(graphene.Mutation):
    """Resumes a set of partition backfill runs. Resuming a backfill will not retry any failed runs."""

    Output = graphene.NonNull(GrapheneResumeBackfillResult)

    class Arguments:
        backfillId = graphene.NonNull(graphene.String)

    class Meta:
        name = "ResumeBackfillMutation"

    @capture_error
    @require_permission_check(Permissions.LAUNCH_PARTITION_BACKFILL)
    def mutate(self, graphene_info: ResolveInfo, backfillId: str):
        return resume_partition_backfill(graphene_info, backfillId)


class GrapheneReexecuteBackfillMutation(graphene.Mutation):
    """Retries a set of partition backfill runs. Retrying a backfill will create a new backfill to retry any failed partitions."""

    Output = graphene.NonNull(GrapheneLaunchBackfillResult)

    class Arguments:
        reexecutionParams = graphene.Argument(GrapheneReexecutionParams)

    class Meta:
        name = "RetryBackfillMutation"

    @capture_error
    @require_permission_check(Permissions.LAUNCH_PARTITION_BACKFILL)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        reexecutionParams: GrapheneReexecutionParams,
    ):
        return retry_partition_backfill(
            graphene_info,
            backfill_id=reexecutionParams["parentRunId"],
            strategy=reexecutionParams["strategy"],
        )


class GrapheneAddDynamicPartitionMutation(graphene.Mutation):
    """Adds a partition to a dynamic partition set."""

    Output = graphene.NonNull(GrapheneAddDynamicPartitionResult)

    class Arguments:
        repositorySelector = graphene.NonNull(GrapheneRepositorySelector)
        partitionsDefName = graphene.NonNull(graphene.String)
        partitionKey = graphene.NonNull(graphene.String)

    class Meta:
        name = "AddDynamicPartitionMutation"

    @capture_error
    @require_permission_check(Permissions.EDIT_DYNAMIC_PARTITIONS)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        repositorySelector: GrapheneRepositorySelector,
        partitionsDefName: str,
        partitionKey: str,
    ):
        return add_dynamic_partition(
            graphene_info, repositorySelector, partitionsDefName, partitionKey
        )


class GrapheneDeleteDynamicPartitionsMutation(graphene.Mutation):
    """Deletes partitions from a dynamic partition set."""

    Output = graphene.NonNull(GrapheneDeleteDynamicPartitionsResult)

    class Arguments:
        repositorySelector = graphene.NonNull(GrapheneRepositorySelector)
        partitionsDefName = graphene.NonNull(graphene.String)
        partitionKeys = non_null_list(graphene.String)

    class Meta:
        name = "DeleteDynamicPartitionsMutation"

    @capture_error
    @require_permission_check(Permissions.EDIT_DYNAMIC_PARTITIONS)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        repositorySelector: GrapheneRepositorySelector,
        partitionsDefName: str,
        partitionKeys: Sequence[str],
    ):
        return delete_dynamic_partitions(
            graphene_info, repositorySelector, partitionsDefName, partitionKeys
        )


def create_execution_params_and_launch_pipeline_reexec(graphene_info, execution_params_dict):
    execution_params = create_execution_params(graphene_info, execution_params_dict)
    assert_permission_for_location(
        graphene_info,
        Permissions.LAUNCH_PIPELINE_REEXECUTION,
        execution_params.selector.location_name,
    )
    return launch_pipeline_reexecution(graphene_info, execution_params=execution_params)


class GrapheneLaunchRunReexecutionMutation(graphene.Mutation):
    """Re-executes a job run."""

    Output = graphene.NonNull(GrapheneLaunchRunReexecutionResult)

    class Arguments:
        executionParams = graphene.Argument(GrapheneExecutionParams)
        reexecutionParams = graphene.Argument(GrapheneReexecutionParams)

    class Meta:
        name = "LaunchRunReexecutionMutation"

    @capture_error
    @require_permission_check(Permissions.LAUNCH_PIPELINE_REEXECUTION)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        executionParams: Optional[GrapheneExecutionParams] = None,
        reexecutionParams: Optional[GrapheneReexecutionParams] = None,
    ):
        if bool(executionParams) == bool(reexecutionParams):
            raise DagsterInvariantViolationError(
                "Must only provide one of either executionParams or reexecutionParams"
            )

        if executionParams:
            return create_execution_params_and_launch_pipeline_reexec(
                graphene_info,
                execution_params_dict=executionParams,
            )
        elif reexecutionParams:
            extra_tags = None
            if reexecutionParams.get("extraTags"):
                extra_tags = {t["key"]: t["value"] for t in reexecutionParams["extraTags"]}

            use_parent_run_tags = None
            if reexecutionParams.get("useParentRunTags") is not None:
                use_parent_run_tags = reexecutionParams["useParentRunTags"]

            return launch_reexecution_from_parent_run(
                graphene_info,
                parent_run_id=reexecutionParams["parentRunId"],
                strategy=reexecutionParams["strategy"],
                extra_tags=extra_tags,
                use_parent_run_tags=use_parent_run_tags,
            )
        else:
            check.failed("Unreachable")


class GrapheneTerminateRunPolicy(graphene.Enum):
    """The type of termination policy to use for a run."""

    # Default behavior: Only mark as canceled if the termination is successful, and after all
    # resources performing the execution have been shut down.
    SAFE_TERMINATE = "SAFE_TERMINATE"

    # Immediately mark the run as canceled, whether or not the termination was successful.
    # No guarantee that the execution has actually stopped.
    MARK_AS_CANCELED_IMMEDIATELY = "MARK_AS_CANCELED_IMMEDIATELY"

    class Meta:
        name = "TerminateRunPolicy"


class GrapheneTerminateRunMutation(graphene.Mutation):
    """Terminates a run."""

    Output = graphene.NonNull(GrapheneTerminateRunResult)

    class Arguments:
        runId = graphene.NonNull(graphene.String)
        terminatePolicy = graphene.Argument(GrapheneTerminateRunPolicy)

    class Meta:
        name = "TerminateRunMutation"

    @capture_error
    @require_permission_check(Permissions.TERMINATE_PIPELINE_EXECUTION)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        runId: str,
        terminatePolicy: Optional[GrapheneTerminateRunPolicy] = None,
    ):
        return terminate_pipeline_execution(
            graphene_info,
            runId,
            terminatePolicy or GrapheneTerminateRunPolicy.SAFE_TERMINATE,
        )


class GrapheneTerminateRunsMutation(graphene.Mutation):
    """Terminates a set of runs given their run IDs."""

    Output = graphene.NonNull(GrapheneTerminateRunsResultOrError)

    class Arguments:
        runIds = non_null_list(graphene.String)
        terminatePolicy = graphene.Argument(GrapheneTerminateRunPolicy)

    class Meta:
        name = "TerminateRunsMutation"

    @capture_error
    @require_permission_check(Permissions.TERMINATE_PIPELINE_EXECUTION)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        runIds: Sequence[str],
        terminatePolicy: Optional[GrapheneTerminateRunPolicy] = None,
    ):
        return terminate_pipeline_execution_for_runs(
            graphene_info,
            runIds,
            terminatePolicy or GrapheneTerminateRunPolicy.SAFE_TERMINATE,
        )


class GrapheneReloadRepositoryLocationMutationResult(graphene.Union):
    """The output from reloading a code location server."""

    class Meta:
        types = (
            GrapheneWorkspaceLocationEntry,
            GrapheneReloadNotSupported,
            GrapheneRepositoryLocationNotFound,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "ReloadRepositoryLocationMutationResult"


class GrapheneShutdownRepositoryLocationSuccess(graphene.ObjectType):
    """Output indicating that a code location server was shut down."""

    repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        name = "ShutdownRepositoryLocationSuccess"


class GrapheneShutdownRepositoryLocationMutationResult(graphene.Union):
    """The output from shutting down a code location server."""

    class Meta:
        types = (
            GrapheneShutdownRepositoryLocationSuccess,
            GrapheneRepositoryLocationNotFound,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "ShutdownRepositoryLocationMutationResult"


class GrapheneReloadRepositoryLocationMutation(graphene.Mutation):
    """Reloads a code location server."""

    Output = graphene.NonNull(GrapheneReloadRepositoryLocationMutationResult)

    class Arguments:
        repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        name = "ReloadRepositoryLocationMutation"

    @capture_error
    @require_permission_check(Permissions.RELOAD_REPOSITORY_LOCATION)
    def mutate(
        self, graphene_info: ResolveInfo, repositoryLocationName: str
    ) -> Union[
        GrapheneWorkspaceLocationEntry,
        GrapheneReloadNotSupported,
        GrapheneRepositoryLocationNotFound,
    ]:
        assert_permission_for_location(
            graphene_info, Permissions.RELOAD_REPOSITORY_LOCATION, repositoryLocationName
        )

        if not graphene_info.context.has_code_location_name(repositoryLocationName):
            return GrapheneRepositoryLocationNotFound(repositoryLocationName)

        if not graphene_info.context.is_reload_supported(repositoryLocationName):
            return GrapheneReloadNotSupported(repositoryLocationName)

        # The current workspace context is a WorkspaceRequestContext, which contains a reference to the
        # repository locations that were present in the root IWorkspaceProcessContext the start of the
        # request. Reloading a repository location modifies the IWorkspaceProcessContext, rendeirng
        # our current WorkspaceRequestContext outdated. Therefore, `reload_repository_location` returns
        # an updated WorkspaceRequestContext for us to use.
        new_context = graphene_info.context.reload_code_location(repositoryLocationName)
        return GrapheneWorkspaceLocationEntry(
            check.not_none(new_context.get_location_entry(repositoryLocationName))
        )


class GrapheneShutdownRepositoryLocationMutation(graphene.Mutation):
    """Shuts down a code location server."""

    Output = graphene.NonNull(GrapheneShutdownRepositoryLocationMutationResult)

    class Arguments:
        repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        name = "ShutdownRepositoryLocationMutation"

    @capture_error
    @require_permission_check(Permissions.RELOAD_REPOSITORY_LOCATION)
    def mutate(
        self, graphene_info: ResolveInfo, repositoryLocationName: str
    ) -> Union[GrapheneRepositoryLocationNotFound, GrapheneShutdownRepositoryLocationSuccess]:
        assert_permission_for_location(
            graphene_info, Permissions.RELOAD_REPOSITORY_LOCATION, repositoryLocationName
        )
        if not graphene_info.context.has_code_location_name(repositoryLocationName):
            return GrapheneRepositoryLocationNotFound(repositoryLocationName)

        if not graphene_info.context.is_shutdown_supported(repositoryLocationName):
            raise Exception(
                f"Location {repositoryLocationName} does not support shutting down via GraphQL"
            )

        graphene_info.context.shutdown_code_location(repositoryLocationName)
        return GrapheneShutdownRepositoryLocationSuccess(
            repositoryLocationName=repositoryLocationName
        )


class GrapheneReloadWorkspaceMutationResult(graphene.Union):
    """The output from reloading the workspace."""

    class Meta:
        types = (
            GrapheneWorkspace,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "ReloadWorkspaceMutationResult"


class GrapheneReloadWorkspaceMutation(graphene.Mutation):
    """Reloads the workspace and its code location servers."""

    Output = graphene.NonNull(GrapheneReloadWorkspaceMutationResult)

    class Meta:
        name = "ReloadWorkspaceMutation"

    @capture_error
    @check_permission(Permissions.RELOAD_WORKSPACE)
    def mutate(self, graphene_info: ResolveInfo):
        new_context = graphene_info.context.reload_workspace()
        return fetch_workspace(new_context)


class GrapheneAssetWipeSuccess(graphene.ObjectType):
    """Output indicating that asset history was deleted."""

    assetPartitionRanges = non_null_list(GrapheneAssetPartitionRange)

    class Meta:
        name = "AssetWipeSuccess"


class GrapheneAssetWipeMutationResult(graphene.Union):
    """The output from deleting asset history."""

    class Meta:
        types = (
            GrapheneAssetNotFoundError,
            GrapheneUnauthorizedError,
            GraphenePythonError,
            GrapheneUnsupportedOperationError,
            GrapheneAssetWipeSuccess,
        )
        name = "AssetWipeMutationResult"


class GrapheneAssetWipeMutation(graphene.Mutation):
    """Deletes asset history from storage."""

    Output = graphene.NonNull(GrapheneAssetWipeMutationResult)

    class Arguments:
        assetPartitionRanges = graphene.Argument(non_null_list(GraphenePartitionsByAssetSelector))

    class Meta:
        name = "AssetWipeMutation"

    @capture_error
    @check_permission(Permissions.WIPE_ASSETS)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        assetPartitionRanges: Sequence[GraphenePartitionsByAssetSelector],
    ) -> GrapheneAssetWipeMutationResult:
        normalized_ranges = [
            AssetPartitionWipeRange(
                AssetKey.from_graphql_input(ap.assetKey),
                PartitionKeyRange(start=ap.partitions.range.start, end=ap.partitions.range.end)
                if ap.partitions
                else None,
            )
            for ap in assetPartitionRanges
        ]

        return wipe_assets(graphene_info, normalized_ranges)


class GrapheneReportRunlessAssetEventsSuccess(graphene.ObjectType):
    """Output indicating that runless asset events were reported."""

    assetKey = graphene.NonNull(GrapheneAssetKey)

    class Meta:
        name = "ReportRunlessAssetEventsSuccess"


class GrapheneReportRunlessAssetEventsResult(graphene.Union):
    """The output from reporting runless events."""

    class Meta:
        types = (
            GrapheneUnauthorizedError,
            GraphenePythonError,
            GrapheneReportRunlessAssetEventsSuccess,
        )
        name = "ReportRunlessAssetEventsResult"


class GrapheneReportRunlessAssetEventsMutation(graphene.Mutation):
    """Reports runless events for an asset or a subset of its partitions."""

    Output = graphene.NonNull(GrapheneReportRunlessAssetEventsResult)

    class Arguments:
        eventParams = graphene.Argument(graphene.NonNull(GrapheneReportRunlessAssetEventsParams))

    class Meta:
        name = "ReportRunlessAssetEventsMutation"

    @capture_error
    @require_permission_check(Permissions.REPORT_RUNLESS_ASSET_EVENTS)
    def mutate(
        self, graphene_info: ResolveInfo, eventParams: GrapheneReportRunlessAssetEventsParams
    ):
        event_type = eventParams["eventType"].to_dagster_event_type()
        asset_key = AssetKey.from_graphql_input(eventParams["assetKey"])
        partition_keys = eventParams.get("partitionKeys", None)
        description = eventParams.get("description", None)

        reporting_user_tags = {**graphene_info.context.get_reporting_user_tags()}

        asset_graph = graphene_info.context.asset_graph

        assert_permission_for_asset_graph(
            graphene_info, asset_graph, [asset_key], Permissions.REPORT_RUNLESS_ASSET_EVENTS
        )

        return report_runless_asset_events(
            graphene_info,
            event_type=event_type,
            asset_key=asset_key,
            partition_keys=partition_keys,
            description=description,
            tags=reporting_user_tags,
        )


class GrapheneLogTelemetrySuccess(graphene.ObjectType):
    """Output indicating that telemetry was logged."""

    action = graphene.NonNull(graphene.String)

    class Meta:
        name = "LogTelemetrySuccess"


class GrapheneLogTelemetryMutationResult(graphene.Union):
    """The output from logging telemetry."""

    class Meta:
        types = (
            GrapheneLogTelemetrySuccess,
            GraphenePythonError,
        )
        name = "LogTelemetryMutationResult"


class GrapheneLogTelemetryMutation(graphene.Mutation):
    """Log telemetry about the Dagster instance."""

    Output = graphene.NonNull(GrapheneLogTelemetryMutationResult)

    class Arguments:
        action = graphene.Argument(graphene.NonNull(graphene.String))
        clientTime = graphene.Argument(graphene.NonNull(graphene.String))
        clientId = graphene.Argument(graphene.NonNull(graphene.String))
        metadata = graphene.Argument(graphene.NonNull(graphene.String))

    class Meta:
        name = "LogTelemetryMutation"

    @capture_error
    def mutate(
        self, graphene_info: ResolveInfo, action: str, clientTime: str, clientId: str, metadata: str
    ):
        action = log_ui_telemetry_event(
            graphene_info,
            action=action,
            client_time=clientTime,
            client_id=clientId,
            metadata=metadata,
        )
        return action


class GrapheneSetNuxSeenMutation(graphene.Mutation):
    """Store whether we've shown the nux to any user and they've dismissed or submitted it."""

    Output = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "SetNuxSeenMutation"

    @capture_error
    def mutate(self, _graphene_info):
        set_nux_seen()
        return get_has_seen_nux()


class GrapheneSetAutoMaterializePausedMutation(graphene.Mutation):
    """Toggle asset auto materializing on or off."""

    Output = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "SetAutoMaterializedPausedMutation"

    class Arguments:
        paused = graphene.Argument(graphene.NonNull(graphene.Boolean))

    @capture_error
    @check_permission(Permissions.TOGGLE_AUTO_MATERIALIZE)
    def mutate(self, graphene_info, paused: bool):
        set_auto_materialize_paused(graphene_info.context.instance, paused)
        return paused


class GrapheneSetConcurrencyLimitMutation(graphene.Mutation):
    """Sets the concurrency limit for a given concurrency key."""

    Output = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "SetConcurrencyLimitMutation"

    class Arguments:
        concurrencyKey = graphene.Argument(graphene.NonNull(graphene.String))
        limit = graphene.Argument(graphene.NonNull(graphene.Int))

    @capture_error
    @check_permission(Permissions.EDIT_CONCURRENCY_LIMIT)
    def mutate(self, graphene_info, concurrencyKey: str, limit: int):
        graphene_info.context.instance.event_log_storage.set_concurrency_slots(
            concurrencyKey, limit
        )
        return True


class GrapheneDeleteConcurrencyLimitMutation(graphene.Mutation):
    """Sets the concurrency limit for a given concurrency key."""

    Output = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "DeleteConcurrencyLimitMutation"

    class Arguments:
        concurrencyKey = graphene.Argument(graphene.NonNull(graphene.String))

    @capture_error
    @check_permission(Permissions.EDIT_CONCURRENCY_LIMIT)
    def mutate(self, graphene_info, concurrencyKey: str):
        graphene_info.context.instance.event_log_storage.delete_concurrency_limit(concurrencyKey)
        return True


class GrapheneFreeConcurrencySlotsMutation(graphene.Mutation):
    """Frees concurrency slots."""

    Output = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "FreeConcurrencySlotsMutation"

    class Arguments:
        runId = graphene.Argument(graphene.NonNull(graphene.String))
        stepKey = graphene.Argument(graphene.String)

    @capture_error
    @check_permission(Permissions.EDIT_CONCURRENCY_LIMIT)
    def mutate(self, graphene_info, runId: str, stepKey: Optional[str] = None):
        event_log_storage = graphene_info.context.instance.event_log_storage
        if stepKey:
            event_log_storage.free_concurrency_slot_for_step(runId, stepKey)
        else:
            event_log_storage.free_concurrency_slots_for_run(runId)
        return True


class GrapheneFreeConcurrencySlotsForRunMutation(graphene.Mutation):
    """Frees the concurrency slots occupied by a specific run."""

    Output = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "FreeConcurrencySlotsForRunMutation"

    class Arguments:
        runId = graphene.Argument(graphene.NonNull(graphene.String))

    @capture_error
    @check_permission(Permissions.EDIT_CONCURRENCY_LIMIT)
    def mutate(self, graphene_info, runId: str):
        graphene_info.context.instance.event_log_storage.free_concurrency_slots_for_run(runId)
        return True


class GrapheneMutation(graphene.ObjectType):
    """The root for all mutations to modify data in your Dagster instance."""

    class Meta:
        name = "Mutation"

    launchPipelineExecution = GrapheneLaunchRunMutation.Field()
    launchRun = GrapheneLaunchRunMutation.Field()
    launchMultipleRuns = GrapheneLaunchMultipleRunsMutation.Field()
    launchPipelineReexecution = GrapheneLaunchRunReexecutionMutation.Field()
    launchRunReexecution = GrapheneLaunchRunReexecutionMutation.Field()
    startSchedule = GrapheneStartScheduleMutation.Field()
    stopRunningSchedule = GrapheneStopRunningScheduleMutation.Field()
    resetSchedule = GrapheneResetScheduleMutation.Field()
    startSensor = GrapheneStartSensorMutation.Field()
    setSensorCursor = GrapheneSetSensorCursorMutation.Field()
    stopSensor = GrapheneStopSensorMutation.Field()
    resetSensor = GrapheneResetSensorMutation.Field()
    sensorDryRun = GrapheneSensorDryRunMutation.Field()
    scheduleDryRun = GrapheneScheduleDryRunMutation.Field()
    terminatePipelineExecution = GrapheneTerminateRunMutation.Field()
    terminateRun = GrapheneTerminateRunMutation.Field()
    terminateRuns = GrapheneTerminateRunsMutation.Field()
    deletePipelineRun = GrapheneDeleteRunMutation.Field()
    deleteRun = GrapheneDeleteRunMutation.Field()
    reloadRepositoryLocation = GrapheneReloadRepositoryLocationMutation.Field()
    reloadWorkspace = GrapheneReloadWorkspaceMutation.Field()
    shutdownRepositoryLocation = GrapheneShutdownRepositoryLocationMutation.Field()
    wipeAssets = GrapheneAssetWipeMutation.Field()
    reportRunlessAssetEvents = GrapheneReportRunlessAssetEventsMutation.Field()
    launchPartitionBackfill = GrapheneLaunchBackfillMutation.Field()
    resumePartitionBackfill = GrapheneResumeBackfillMutation.Field()
    reexecutePartitionBackfill = GrapheneReexecuteBackfillMutation.Field()
    cancelPartitionBackfill = GrapheneCancelBackfillMutation.Field()
    logTelemetry = GrapheneLogTelemetryMutation.Field()
    setNuxSeen = GrapheneSetNuxSeenMutation.Field()
    addDynamicPartition = GrapheneAddDynamicPartitionMutation.Field()
    deleteDynamicPartitions = GrapheneDeleteDynamicPartitionsMutation.Field()
    setAutoMaterializePaused = GrapheneSetAutoMaterializePausedMutation.Field()
    setConcurrencyLimit = GrapheneSetConcurrencyLimitMutation.Field()
    deleteConcurrencyLimit = GrapheneDeleteConcurrencyLimitMutation.Field()
    freeConcurrencySlotsForRun = GrapheneFreeConcurrencySlotsForRunMutation.Field()
    freeConcurrencySlots = GrapheneFreeConcurrencySlotsMutation.Field()



from typing import TYPE_CHECKING, Optional, Union

import dagster._check as check
from dagster._core.definitions.instigation_logger import get_instigation_log_records
from dagster._core.definitions.selector import InstigatorSelector
from dagster._core.log_manager import LOG_RECORD_METADATA_ATTR
from dagster._core.remote_representation.external import CompoundID
from dagster._core.scheduler.instigation import InstigatorStatus

if TYPE_CHECKING:
    from dagster_graphql.schema.instigation import (
        GrapheneInstigationEventConnection,
        GrapheneInstigationState,
        GrapheneInstigationStateNotFoundError,
    )
    from dagster_graphql.schema.util import ResolveInfo


def get_instigator_state_by_selector(
    graphene_info: "ResolveInfo",
    selector: InstigatorSelector,
    instigator_id: Optional[CompoundID],
) -> Union["GrapheneInstigationState", "GrapheneInstigationStateNotFoundError"]:
    from dagster_graphql.schema.instigation import (
        GrapheneInstigationState,
        GrapheneInstigationStateNotFoundError,
    )

    check.inst_param(selector, "selector", InstigatorSelector)

    if instigator_id:
        state = graphene_info.context.instance.get_instigator_state(
            origin_id=instigator_id.remote_origin_id,
            selector_id=instigator_id.selector_id,
        )
        # if the state tells us the status on its own short cut and return it
        # if its declared in code we need the full snapshot to resolve
        if state and state.status in (InstigatorStatus.STOPPED, InstigatorStatus.RUNNING):
            return GrapheneInstigationState(state)

    location = graphene_info.context.get_code_location(selector.location_name)
    repository = location.get_repository(selector.repository_name)

    if repository.has_sensor(selector.name):
        sensor = repository.get_sensor(selector.name)
        stored_state = graphene_info.context.instance.get_instigator_state(
            sensor.get_remote_origin_id(),
            sensor.selector_id,
        )
        current_state = sensor.get_current_instigator_state(stored_state)
    elif repository.has_schedule(selector.name):
        schedule = repository.get_schedule(selector.name)
        stored_state = graphene_info.context.instance.get_instigator_state(
            schedule.get_remote_origin_id(),
            schedule.selector_id,
        )
        current_state = schedule.get_current_instigator_state(stored_state)
    else:
        return GrapheneInstigationStateNotFoundError(selector.name)

    return GrapheneInstigationState(current_state)


def get_instigation_states_by_repository_id(
    graphene_info: "ResolveInfo",
    repository_id: CompoundID,
):
    from dagster_graphql.schema.instigation import (
        GrapheneInstigationState,
        GrapheneInstigationStates,
    )

    states = graphene_info.context.instance.all_instigator_state(
        repository_origin_id=repository_id.remote_origin_id,
        repository_selector_id=repository_id.selector_id,
    )

    return GrapheneInstigationStates(results=[GrapheneInstigationState(state) for state in states])


def get_tick_log_events(graphene_info: "ResolveInfo", tick) -> "GrapheneInstigationEventConnection":
    from dagster_graphql.schema.instigation import (
        GrapheneInstigationEvent,
        GrapheneInstigationEventConnection,
    )
    from dagster_graphql.schema.logs.log_level import GrapheneLogLevel

    if not tick.log_key:
        return GrapheneInstigationEventConnection(events=[], cursor="", hasMore=False)

    records = get_instigation_log_records(graphene_info.context.instance, tick.log_key)

    events = []
    for record_dict in records:
        exc_info = record_dict.get("exc_info")
        message = record_dict[LOG_RECORD_METADATA_ATTR]["orig_message"]
        if exc_info:
            message = f"{message}\n\n{exc_info}"

        event = GrapheneInstigationEvent(
            message=message,
            level=GrapheneLogLevel.from_level(record_dict["levelno"]),
            timestamp=int(record_dict["created"] * 1000),
        )

        events.append(event)

    return GrapheneInstigationEventConnection(
        events=events,
        cursor=None,
        hasMore=False,
    )



import asyncio
import sys
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Optional, Union, cast

import dagster._check as check
import graphene
from dagster._core.definitions.asset_job import IMPLICIT_ASSET_JOB_NAME
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation.external import RemoteRepository
from dagster._core.test_utils import wait_for_runs_to_finish
from dagster._core.workspace.context import WorkspaceProcessContext, WorkspaceRequestContext
from dagster._core.workspace.load_target import PythonFileTarget
from typing_extensions import Protocol, TypeAlias, TypedDict

from dagster_graphql import __file__ as dagster_graphql_init_py
from dagster_graphql.schema import create_schema


class GqlResult(Protocol):
    @property
    def data(self) -> Mapping[str, Any]: ...

    @property
    def errors(self) -> Optional[Sequence[str]]: ...


Selector: TypeAlias = dict[str, Any]

GqlVariables: TypeAlias = Mapping[str, Any]


class GqlTag(TypedDict):
    key: str
    value: str


class GqlAssetKey(TypedDict):
    path: Sequence[str]


class GqlAssetCheckHandle(TypedDict):
    assetKey: GqlAssetKey
    name: str


def main_repo_location_name() -> str:
    return "test_location"


def main_repo_name() -> str:
    return "test_repo"


SCHEMA = create_schema()


def execute_dagster_graphql(
    context: WorkspaceRequestContext,
    query: str,
    variables: Optional[GqlVariables] = None,
    schema: graphene.Schema = SCHEMA,
) -> GqlResult:
    result = asyncio.run(
        schema.execute_async(
            query,
            context_value=context,
            variable_values=variables,
        )
    )
    # It would be cleaner if we instead passed in a process context
    # and made a request context for this invocation.
    # For now just ensure we don't shared loaders between requests.
    context.loaders.clear()

    if result.errors:
        first_error = result.errors[0]
        if hasattr(first_error, "original_error") and first_error.original_error:
            raise result.errors[0].original_error

        raise result.errors[0]

    return result


def execute_dagster_graphql_subscription(
    context: WorkspaceRequestContext,
    query: str,
    variables: Optional[GqlVariables] = None,
    schema: graphene.Schema = SCHEMA,
) -> Sequence[GqlResult]:
    results = []

    subscription = schema.subscribe(
        query,
        context_value=context,
        variable_values=variables,
    )

    async def _process():
        payload_aiter = await subscription
        async for res in payload_aiter:
            results.append(res)
            # first payload should have it all
            break

    asyncio.run(_process())

    return results


def execute_dagster_graphql_and_finish_runs(
    context: WorkspaceRequestContext, query: str, variables: Optional[GqlVariables] = None
) -> GqlResult:
    result = execute_dagster_graphql(context, query, variables)
    wait_for_runs_to_finish(context.instance, timeout=30)
    return result


@contextmanager
def define_out_of_process_context(
    python_file: str,
    fn_name: Optional[str],
    instance: DagsterInstance,
    read_only: bool = False,
    read_only_locations: Optional[Mapping[str, bool]] = None,
) -> Iterator[WorkspaceRequestContext]:
    check.inst_param(instance, "instance", DagsterInstance)

    with define_out_of_process_workspace(
        python_file, fn_name, instance, read_only=read_only
    ) as workspace_process_context:
        yield WorkspaceRequestContext(
            instance=instance,
            current_workspace=workspace_process_context.get_current_workspace(),
            process_context=workspace_process_context,
            version=workspace_process_context.version,
            source=None,
            read_only=read_only,
            read_only_locations=read_only_locations,
        )


def define_out_of_process_workspace(
    python_file: str, fn_name: Optional[str], instance: DagsterInstance, read_only: bool = False
) -> WorkspaceProcessContext:
    return WorkspaceProcessContext(
        instance,
        PythonFileTarget(
            python_file=python_file,
            attribute=fn_name,
            working_directory=None,
            location_name=main_repo_location_name(),
        ),
        version="",
        read_only=read_only,
    )


def infer_repository(graphql_context: WorkspaceRequestContext) -> RemoteRepository:
    if len(graphql_context.code_locations) == 1:
        # This is to account for having a single in process repository
        code_location = graphql_context.code_locations[0]
        repositories = code_location.get_repositories()
        assert len(repositories) == 1
        return next(iter(repositories.values()))

    code_location = graphql_context.get_code_location(main_repo_location_name())
    return code_location.get_repository("test_repo")


def infer_repository_selector(graphql_context: WorkspaceRequestContext) -> Selector:
    if len(graphql_context.code_locations) == 1:
        # This is to account for having a single in process repository
        code_location = graphql_context.code_locations[0]
        repositories = code_location.get_repositories()
        assert len(repositories) == 1
        repository = next(iter(repositories.values()))
    else:
        code_location = graphql_context.get_code_location(main_repo_location_name())
        repository = code_location.get_repository("test_repo")

    return {
        "repositoryLocationName": code_location.name,
        "repositoryName": repository.name,
    }


def infer_job_selector(
    graphql_context: WorkspaceRequestContext,
    job_name: str,
    op_selection: Optional[Sequence[str]] = None,
    asset_selection: Optional[Sequence[GqlAssetKey]] = None,
    asset_check_selection: Optional[Sequence[GqlAssetCheckHandle]] = None,
) -> Selector:
    selector = infer_repository_selector(graphql_context)
    selector.update(
        {
            "pipelineName": job_name,
            "solidSelection": op_selection,
            "assetSelection": asset_selection,
            "assetCheckSelection": asset_check_selection,
        }
    )
    return selector


def infer_schedule_selector(
    graphql_context: WorkspaceRequestContext, schedule_name: str
) -> Selector:
    selector = infer_repository_selector(graphql_context)
    selector.update({"scheduleName": schedule_name})
    return selector


def infer_sensor_selector(graphql_context: WorkspaceRequestContext, sensor_name: str) -> Selector:
    selector = infer_repository_selector(graphql_context)
    selector.update({"sensorName": sensor_name})
    return selector


def infer_instigation_selector(graphql_context: WorkspaceRequestContext, name: str) -> Selector:
    selector = infer_repository_selector(graphql_context)
    selector.update({"name": name})
    return selector


def infer_resource_selector(graphql_context: WorkspaceRequestContext, name: str) -> Selector:
    selector = infer_repository_selector(graphql_context)
    selector = {**selector, **{"resourceName": name}}
    return selector


def ensure_dagster_graphql_tests_import() -> None:
    dagster_package_root = (Path(dagster_graphql_init_py) / ".." / "..").resolve()
    assert (dagster_package_root / "dagster_graphql_tests").exists(), (
        "Could not find dagster_graphql_tests where expected"
    )
    sys.path.append(dagster_package_root.as_posix())


def materialize_assets(
    context: WorkspaceRequestContext,
    asset_selection: Optional[Sequence[AssetKey]] = None,
    partition_keys: Optional[Sequence[str]] = None,
    run_config_data: Optional[Mapping[str, Any]] = None,
) -> Union[GqlResult, Sequence[GqlResult]]:
    from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION

    gql_asset_selection = (
        cast("Sequence[GqlAssetKey]", [key.to_graphql_input() for key in asset_selection])
        if asset_selection
        else None
    )
    selector = infer_job_selector(
        context, IMPLICIT_ASSET_JOB_NAME, asset_selection=gql_asset_selection
    )
    if partition_keys:
        results = []
        for key in partition_keys:
            results.append(
                execute_dagster_graphql(
                    context,
                    LAUNCH_PIPELINE_EXECUTION_MUTATION,
                    variables={
                        "executionParams": {
                            "selector": selector,
                            "executionMetadata": {
                                "tags": [{"key": "dagster/partition", "value": key}]
                            },
                            "runConfigData": run_config_data,
                        }
                    },
                )
            )
        return results
    else:
        selector = infer_job_selector(
            context, IMPLICIT_ASSET_JOB_NAME, asset_selection=gql_asset_selection
        )
        return execute_dagster_graphql(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": run_config_data,
                }
            },
        )



from typing import Optional

import dagster._check as check
import graphene
from dagster._core.definitions.events import AssetKey
from dagster._utils.error import SerializableErrorInfo

from dagster_graphql.schema.util import ResolveInfo, non_null_list


class GrapheneError(graphene.Interface):
    message = graphene.String(required=True)

    class Meta:
        name = "Error"


class GrapheneErrorChainLink(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ErrorChainLink"

    error = graphene.NonNull(lambda: GraphenePythonError)
    isExplicitLink = graphene.NonNull(graphene.Boolean)


class GraphenePythonError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PythonError"

    className = graphene.Field(graphene.String)
    stack = non_null_list(graphene.String)
    cause = graphene.Field(lambda: GraphenePythonError)

    """
    A list of all recursive errors deeper in the exception stack that caused this error to be
    raised, only including explicit raises. For example, in a block of code like:
    ```
        try:
            try:
                raise Exception("Inner")
            except Exception as e:
                raise Exception("Middle") from e
        except Exception as e:
            raise Exception("Outer") from e
    ```
    The PythonError returned will correspond to Outer, with two causes - Middle and then Inner
    """
    causes = non_null_list("dagster_graphql.schema.errors.GraphenePythonError")

    """
    A list of all recursive errors deeper in the exception stack that caused this error to be
    raised, including both explicit and implicit links. For example, in a block of code like:
    ```
        try:
            try:
                raise Exception("Inner")
            except Exception as e:
                raise Exception("Middle")
        except Exception as e:
            raise Exception("Outer") from e
    ```
    The PythonError returned will correspond to Outer, with two links in the chain - one explicit (Middle) then one implicit (Inner)
    """

    errorChain = non_null_list(GrapheneErrorChainLink)

    def __init__(self, error_info):
        super().__init__()
        self._error_info = check.inst_param(error_info, "error_info", SerializableErrorInfo)
        self._message = error_info.message
        self._stack = error_info.stack
        self._className = error_info.cls_name
        self._cause = error_info.cause
        self._context = error_info.context

    def resolve_message(self, _graphene_info):
        check.invariant(
            isinstance(self, GraphenePythonError),
            f"GraphenePythonError methods called on a {type(self)} - this usually indicates"
            " that a SerializableErrorInfo was passed in where a GraphenePythonError was"
            " expected",
        )
        return self._message

    def resolve_stack(self, _graphene_info):
        check.invariant(
            isinstance(self, GraphenePythonError),
            f"GraphenePythonError methods called on a {type(self)} - this usually indicates"
            " that a SerializableErrorInfo was passed in where a GraphenePythonError was"
            " expected",
        )
        return self._stack

    def resolve_cause(self, _graphene_info):
        return GraphenePythonError(self._cause) if self._cause else None

    def resolution_context(self, _graphene_info):
        return GraphenePythonError(self._context) if self._context else None

    def resolve_className(self, _graphene_info):
        check.invariant(
            isinstance(self, GraphenePythonError),
            f"GraphenePythonError methods called on a {type(self)} - this usually indicates"
            " that a SerializableErrorInfo was passed in where a GraphenePythonError was"
            " expected",
        )
        return self._className

    def resolve_causes(self, _graphene_info: ResolveInfo):
        causes: list[GraphenePythonError] = []
        current_error = self._cause
        while current_error and len(causes) < 10:  # Sanity check the depth of the causes
            causes.append(GraphenePythonError(current_error))
            current_error = current_error.cause
        return causes

    def resolve_errorChain(self, _graphene_info):
        current_error = self._error_info
        chain = []
        while len(chain) < 10:  # Sanity check the length of the chain
            if current_error.cause:
                current_error = current_error.cause
                chain.append(
                    GrapheneErrorChainLink(
                        error=GraphenePythonError(current_error), isExplicitLink=True
                    )
                )
            elif current_error.context:
                current_error = current_error.context
                chain.append(
                    GrapheneErrorChainLink(
                        error=GraphenePythonError(current_error), isExplicitLink=False
                    )
                )
            else:
                break
        return chain


class GrapheneSchedulerNotDefinedError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "SchedulerNotDefinedError"

    def __init__(self):
        super().__init__()
        self.message = "Scheduler is not defined for the currently loaded repository."


class GraphenePipelineSnapshotNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PipelineSnapshotNotFoundError"

    snapshot_id = graphene.NonNull(graphene.String)

    def __init__(self, snapshot_id):
        super().__init__()
        self.snapshot_id = check.str_param(snapshot_id, "snapshot_id")
        self.message = f"Pipeline snapshot {snapshot_id} is not present in the current instance."


class GrapheneReloadNotSupported(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ReloadNotSupported"

    def __init__(self, location_name):
        super().__init__()
        self.message = f"Location {location_name} does not support reloading."


class GrapheneRepositoryLocationNotFound(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "RepositoryLocationNotFound"

    def __init__(self, location_name):
        super().__init__()
        self.message = f"Location {location_name} does not exist."


class GraphenePipelineNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PipelineNotFoundError"

    pipeline_name = graphene.NonNull(graphene.String)
    repository_name = graphene.NonNull(graphene.String)
    repository_location_name = graphene.NonNull(graphene.String)

    def __init__(self, selector):
        from dagster_graphql.implementation.utils import JobSubsetSelector

        super().__init__()
        check.inst_param(selector, "selector", JobSubsetSelector)
        self.pipeline_name = selector.job_name
        self.repository_name = selector.repository_name
        self.repository_location_name = selector.location_name
        self.message = (
            "Could not find Pipeline "
            f"{selector.location_name}.{selector.repository_name}.{selector.job_name}"
        )


class GrapheneGraphNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "GraphNotFoundError"

    graph_name = graphene.NonNull(graphene.String)
    repository_name = graphene.NonNull(graphene.String)
    repository_location_name = graphene.NonNull(graphene.String)

    def __init__(self, selector):
        from dagster_graphql.implementation.utils import GraphSelector

        super().__init__()
        check.inst_param(selector, "selector", GraphSelector)
        self.graph_name = selector.graph_name
        self.repository_name = selector.repository_name
        self.repository_location_name = selector.location_name
        self.message = (
            "Could not find Graph "
            f"{selector.location_name}.{selector.repository_name}.{selector.graph_name}"
        )


class GraphenePipelineRunNotFoundError(graphene.Interface):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PipelineRunNotFoundError"

    run_id = graphene.NonNull(graphene.String)
    message = graphene.String(required=True)


class GrapheneRunNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GraphenePipelineRunNotFoundError, GrapheneError)
        name = "RunNotFoundError"

    run_id = graphene.NonNull(graphene.String)

    def __init__(self, run_id):
        super().__init__()
        self.run_id = check.str_param(run_id, "run_id")
        self.message = f"Pipeline run {run_id} could not be found."


class GrapheneInvalidPipelineRunsFilterError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "InvalidPipelineRunsFilterError"

    def __init__(self, message):
        super().__init__()
        self.message = check.str_param(message, "message")


class GrapheneRunGroupNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "RunGroupNotFoundError"

    run_id = graphene.NonNull(graphene.String)

    def __init__(self, run_id):
        super().__init__()
        self.run_id = check.str_param(run_id, "run_id")
        self.message = f"Run group of run {run_id} could not be found."


class GraphenePresetNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PresetNotFoundError"

    preset = graphene.NonNull(graphene.String)

    def __init__(self, preset, selector):
        super().__init__()
        self.preset = check.str_param(preset, "preset")
        self.message = f"Preset {preset} not found in pipeline {selector.job_name}."


class GrapheneConflictingExecutionParamsError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ConflictingExecutionParamsError"

    def __init__(self, conflicting_param):
        super().__init__()
        self.message = (
            f"Invalid ExecutionParams. Cannot define {conflicting_param} when using a preset."
        )


class GrapheneModeNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ModeNotFoundError"

    mode = graphene.NonNull(graphene.String)

    def __init__(self, mode, selector):
        super().__init__()
        self.mode = check.str_param(mode, "mode")
        self.message = f"Mode {mode} not found in pipeline {selector.job_name}."


class GrapheneNoModeProvidedError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "NoModeProvidedError"

    pipeline_name = graphene.NonNull(graphene.String)

    def __init__(self, pipeline_name, mode_list):
        super().__init__()
        mode_list = check.list_param(mode_list, "mode_list", of_type=str)
        pipeline_name = check.str_param(pipeline_name, "pipeline_name")
        self.message = (
            f"No mode provided for pipeline '{pipeline_name}', which has multiple modes. Available"
            f" modes: {mode_list}"
        )


class GrapheneInvalidStepError(graphene.ObjectType):
    invalid_step_key = graphene.NonNull(graphene.String)

    class Meta:
        name = "InvalidStepError"


class GrapheneInvalidOutputError(graphene.ObjectType):
    step_key = graphene.NonNull(graphene.String)
    invalid_output_name = graphene.NonNull(graphene.String)

    class Meta:
        name = "InvalidOutputError"


class GraphenePipelineRunConflict(graphene.Interface):
    message = graphene.NonNull(graphene.String)

    class Meta:
        name = "PipelineRunConflict"


class GrapheneRunConflict(graphene.ObjectType):
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneError, GraphenePipelineRunConflict)
        name = "RunConflict"


class GrapheneInvalidSubsetError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "InvalidSubsetError"

    pipeline = graphene.Field(
        graphene.NonNull("dagster_graphql.schema.pipelines.pipeline.GraphenePipeline")
    )

    def __init__(self, message, pipeline):
        super().__init__()
        self.message = check.str_param(message, "message")
        self.pipeline = pipeline


class GrapheneConfigTypeNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ConfigTypeNotFoundError"

    pipeline = graphene.Field(
        graphene.NonNull("dagster_graphql.schema.pipelines.pipeline.GraphenePipeline")
    )
    config_type_name = graphene.NonNull(graphene.String)


create_execution_params_error_types = (
    GraphenePresetNotFoundError,
    GrapheneConflictingExecutionParamsError,
    GrapheneNoModeProvidedError,
)


class GrapheneDagsterTypeNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "DagsterTypeNotFoundError"

    dagster_type_name = graphene.NonNull(graphene.String)


class GrapheneScheduleNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ScheduleNotFoundError"

    schedule_name = graphene.NonNull(graphene.String)

    def __init__(self, schedule_name):
        super().__init__()
        self.schedule_name = check.str_param(schedule_name, "schedule_name")
        self.message = f"Schedule {self.schedule_name} could not be found."


class GrapheneResourceNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ResourceNotFoundError"

    resource_name = graphene.NonNull(graphene.String)

    def __init__(self, resource_name):
        super().__init__()
        self.resource_name = check.str_param(resource_name, "resource_name")
        self.message = f"Top-level resource {self.resource_name} could not be found."


class GrapheneSensorNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "SensorNotFoundError"

    sensor_name = graphene.NonNull(graphene.String)

    def __init__(self, sensor_name):
        super().__init__()
        self.sensor_name = check.str_param(sensor_name, "sensor_name")
        self.message = f"Could not find `{sensor_name}` in the currently loaded repository."


class GraphenePartitionSetNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PartitionSetNotFoundError"

    partition_set_name = graphene.NonNull(graphene.String)

    def __init__(self, partition_set_name):
        super().__init__()
        self.partition_set_name = check.str_param(partition_set_name, "partition_set_name")
        self.message = f"Partition set {self.partition_set_name} could not be found."


class GraphenePartitionKeysNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PartitionKeysNotFoundError"

    partition_keys = non_null_list(graphene.String)

    def __init__(self, partition_keys: set[str]):
        super().__init__()
        self.partition_keys = check.list_param(
            sorted(partition_keys), "partition_keys", of_type=str
        )
        self.message = f"Partition keys `{self.partition_keys}` could not be found."


class GrapheneRepositoryNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "RepositoryNotFoundError"

    repository_name = graphene.NonNull(graphene.String)
    repository_location_name = graphene.NonNull(graphene.String)

    def __init__(self, repository_location_name, repository_name):
        super().__init__()
        self.repository_name = check.str_param(repository_name, "repository_name")
        self.repository_location_name = check.str_param(
            repository_location_name, "repository_location_name"
        )
        self.message = f"Could not find Repository {repository_location_name}.{repository_name}"


class GrapheneAssetNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "AssetNotFoundError"

    def __init__(self, asset_key):
        super().__init__()
        self.asset_key = check.inst_param(asset_key, "asset_key", AssetKey)
        self.message = f"Asset key {asset_key.to_string()} not found."


class GrapheneUnauthorizedError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "UnauthorizedError"

    def __init__(self, message=None):
        super().__init__()
        self.message = message if message else "Authorization failed"


class GrapheneDuplicateDynamicPartitionError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "DuplicateDynamicPartitionError"

    partitions_def_name = graphene.NonNull(graphene.String)
    partition_name = graphene.NonNull(graphene.String)

    def __init__(self, partitions_def_name, partition_name):
        super().__init__()
        self.partitions_def_name = check.str_param(partitions_def_name, "partitions_def_name")
        self.partition_name = check.str_param(partition_name, "partition_name")
        self.message = (
            f"Partition {self.partition_name} already exists in dynamic partitions definition"
            f" {self.partitions_def_name}."
        )


class GrapheneUnsupportedOperationError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "UnsupportedOperationError"

    def __init__(self, message: Optional[str] = None):
        super().__init__()
        self.message = check.str_param(message, "message") or "Unsupported operation."


types = [
    GrapheneAssetNotFoundError,
    GrapheneConflictingExecutionParamsError,
    GrapheneConfigTypeNotFoundError,
    GrapheneDagsterTypeNotFoundError,
    GrapheneError,
    GrapheneInvalidOutputError,
    GrapheneInvalidPipelineRunsFilterError,
    GrapheneInvalidStepError,
    GrapheneInvalidSubsetError,
    GrapheneModeNotFoundError,
    GrapheneNoModeProvidedError,
    GraphenePartitionSetNotFoundError,
    GraphenePipelineNotFoundError,
    GraphenePipelineRunConflict,
    GrapheneRunConflict,
    GraphenePipelineRunNotFoundError,
    GraphenePipelineSnapshotNotFoundError,
    GraphenePresetNotFoundError,
    GraphenePythonError,
    GrapheneUnauthorizedError,
    GrapheneReloadNotSupported,
    GrapheneRepositoryLocationNotFound,
    GrapheneRepositoryNotFoundError,
    GrapheneResourceNotFoundError,
    GrapheneRunGroupNotFoundError,
    GrapheneRunNotFoundError,
    GrapheneScheduleNotFoundError,
    GrapheneSchedulerNotDefinedError,
    GrapheneSensorNotFoundError,
    GrapheneUnsupportedOperationError,
    GrapheneDuplicateDynamicPartitionError,
]



from typing import Callable

import dagster._check as check
import graphene
from dagster._config.snap import ConfigTypeSnap
from dagster._core.snap import ResourceDefSnap

from dagster_graphql.schema.config_types import GrapheneConfigTypeField
from dagster_graphql.schema.errors import (
    GrapheneInvalidSubsetError,
    GraphenePipelineNotFoundError,
    GraphenePythonError,
)
from dagster_graphql.schema.util import ResolveInfo, non_null_list


class GrapheneResource(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    configField = graphene.Field(GrapheneConfigTypeField)

    class Meta:
        name = "Resource"

    def __init__(
        self,
        get_config_type: Callable[[str], ConfigTypeSnap],
        resource_def_snap: ResourceDefSnap,
    ):
        super().__init__()
        self._get_config_type = get_config_type
        self._resource_def_snap = check.inst_param(
            resource_def_snap, "resource_def_snap", ResourceDefSnap
        )
        self.name = resource_def_snap.name
        self.description = resource_def_snap.description

    def resolve_configField(self, _graphene_info: ResolveInfo):
        if self._resource_def_snap.config_field_snap:
            try:
                # config type may not be present if mode config mapped, null out gracefully
                self._get_config_type(self._resource_def_snap.config_field_snap.type_key)
            except KeyError:
                return None

            return GrapheneConfigTypeField(
                get_config_type=self._get_config_type,
                field_snap=self._resource_def_snap.config_field_snap,
            )

        return None


class GrapheneResourceConnection(graphene.ObjectType):
    class Meta:
        name = "ResourceConnection"

    resources = non_null_list(GrapheneResource)


class GrapheneResourcesOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneResourceConnection,
            GraphenePipelineNotFoundError,
            GrapheneInvalidSubsetError,
            GraphenePythonError,
        )
        name = "ResourcesOrError"



import graphene

from dagster_graphql.schema.errors import GrapheneAssetNotFoundError, GraphenePythonError
from dagster_graphql.schema.pipelines.pipeline import GrapheneAsset, GrapheneAssetRecord
from dagster_graphql.schema.util import non_null_list


class GrapheneAssetConnection(graphene.ObjectType):
    nodes = non_null_list(GrapheneAsset)
    cursor = graphene.Field(graphene.String)

    class Meta:
        name = "AssetConnection"


class GrapheneAssetRecordConnection(graphene.ObjectType):
    assets = non_null_list(GrapheneAssetRecord)
    cursor = graphene.Field(graphene.String)

    class Meta:
        name = "AssetRecordConnection"


class GrapheneAssetsOrError(graphene.Union):
    class Meta:
        types = (GrapheneAssetConnection, GraphenePythonError)
        name = "AssetsOrError"


class GrapheneAssetRecordsOrError(graphene.Union):
    class Meta:
        types = (GrapheneAssetRecordConnection, GraphenePythonError)
        name = "AssetRecordsOrError"


class GrapheneAssetOrError(graphene.Union):
    class Meta:
        types = (GrapheneAsset, GrapheneAssetNotFoundError)
        name = "AssetOrError"



import graphene

from dagster_graphql.schema.errors import (
    GrapheneGraphNotFoundError,
    GrapheneInvalidSubsetError,
    GraphenePipelineNotFoundError,
    GraphenePythonError,
)
from dagster_graphql.schema.pipelines.pipeline import GrapheneGraph, GraphenePipeline


class GraphenePipelineOrError(graphene.Union):
    class Meta:
        types = (
            GraphenePipeline,
            GraphenePipelineNotFoundError,
            GrapheneInvalidSubsetError,
            GraphenePythonError,
        )
        name = "PipelineOrError"


class GrapheneGraphOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneGraph,
            GrapheneGraphNotFoundError,
            GraphenePythonError,
        )
        name = "GraphOrError"



import asyncio
from typing import Optional

import graphene
from dagster import _check as check
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.freshness import FreshnessState
from dagster._core.storage.asset_check_execution_record import AssetCheckExecutionResolvedStatus
from dagster._core.storage.dagster_run import RunRecord
from dagster._core.storage.event_log.base import AssetCheckSummaryRecord, AssetRecord

from dagster_graphql.implementation.fetch_partition_subsets import (
    regenerate_and_check_partition_subsets,
)
from dagster_graphql.schema.util import ResolveInfo


class GrapheneAssetHealthStatus(graphene.Enum):
    HEALTHY = "HEALTHY"
    WARNING = "WARNING"
    DEGRADED = "DEGRADED"
    UNKNOWN = "UNKNOWN"
    NOT_APPLICABLE = "NOT_APPLICABLE"

    class Meta:
        name = "AssetHealthStatus"


class GrapheneAssetHealthCheckDegradedMeta(graphene.ObjectType):
    numFailedChecks = graphene.NonNull(graphene.Int)
    numWarningChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckDegradedMeta"


class GrapheneAssetHealthCheckWarningMeta(graphene.ObjectType):
    numWarningChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckWarningMeta"


class GrapheneAssetHealthCheckUnknownMeta(graphene.ObjectType):
    numNotExecutedChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckUnknownMeta"


class GrapheneAssetHealthCheckMeta(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetHealthCheckDegradedMeta,
            GrapheneAssetHealthCheckWarningMeta,
            GrapheneAssetHealthCheckUnknownMeta,
        )
        name = "AssetHealthCheckMeta"


class GrapheneAssetHealthMaterializationDegradedPartitionedMeta(graphene.ObjectType):
    numFailedPartitions = graphene.NonNull(graphene.Int)
    numMissingPartitions = graphene.NonNull(graphene.Int)
    totalNumPartitions = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthMaterializationDegradedPartitionedMeta"


class GrapheneAssetHealthMaterializationHealthyPartitionedMeta(graphene.ObjectType):
    numMissingPartitions = graphene.NonNull(graphene.Int)
    totalNumPartitions = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthMaterializationHealthyPartitionedMeta"


class GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta(graphene.ObjectType):
    failedRunId = graphene.NonNull(graphene.String)

    class Meta:
        name = "AssetHealthMaterializationDegradedNotPartitionedMeta"


class GrapheneAssetHealthMaterializationMeta(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetHealthMaterializationDegradedPartitionedMeta,
            GrapheneAssetHealthMaterializationHealthyPartitionedMeta,
            GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta,
        )
        name = "AssetHealthMaterializationMeta"


class GrapheneAssetHealthFreshnessMeta(graphene.ObjectType):
    lastMaterializedTimestamp = graphene.Field(graphene.Float)

    class Meta:
        name = "AssetHealthFreshnessMeta"


class GrapheneAssetHealth(graphene.ObjectType):
    assetHealth = graphene.NonNull(GrapheneAssetHealthStatus)
    materializationStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    materializationStatusMetadata = graphene.Field(GrapheneAssetHealthMaterializationMeta)
    assetChecksStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    assetChecksStatusMetadata = graphene.Field(GrapheneAssetHealthCheckMeta)
    freshnessStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    freshnessStatusMetadata = graphene.Field(GrapheneAssetHealthFreshnessMeta)

    class Meta:
        name = "AssetHealth"

    def __init__(self, asset_node_snap, dynamic_partitions_loader):
        super().__init__()
        self._asset_node_snap = asset_node_snap
        self._dynamic_partitions_loader = dynamic_partitions_loader
        self.materialization_status_task = None
        self.asset_check_status_task = None
        self.freshness_status_task = None

    async def get_materialization_status_for_asset_health(
        self, graphene_info: ResolveInfo
    ) -> tuple[str, Optional[GrapheneAssetHealthMaterializationMeta]]:
        """Computes the health indicator for the asset materialization status. Follows these rules:
        If the asset is partitioned:
            - HEALTHY - all partitions successfully materialized.
            - WARNING - some partitions are successful but any number of partitions are missing (but no failed partitions).
            - DEGRADED - any number of partitions are failed.
            - UNKNOWN - all partitions are missing.
        If the asset is not partitioned:
            - HEALTHY - the latest materialization of an asset was successfully materialized.
            - WARNING - no conditions lead to a warning status.
            - DEGRADED - latest materialization of an asset failed.
            - UNKNOWN - asset has never had a materialization attempt.
        """
        partitions_snap = self._asset_node_snap.partitions
        asset_key = self._asset_node_snap.asset_key
        if partitions_snap is not None:  # isPartitioned
            (
                materialized_partition_subset,
                failed_partition_subset,
                _,
            ) = regenerate_and_check_partition_subsets(
                graphene_info.context, self._asset_node_snap, self._dynamic_partitions_loader
            )
            total_num_partitions = partitions_snap.get_partitions_definition().get_num_partitions(
                dynamic_partitions_store=self._dynamic_partitions_loader
            )
            currently_materialized_subset = materialized_partition_subset - failed_partition_subset
            num_materialized = len(currently_materialized_subset)
            num_failed = len(failed_partition_subset)
            if num_materialized == 0 and num_failed == 0:
                # asset has never been materialized
                return GrapheneAssetHealthStatus.UNKNOWN, None
            num_missing = total_num_partitions - num_materialized - num_failed
            if num_failed > 0:
                return (
                    GrapheneAssetHealthStatus.DEGRADED,
                    GrapheneAssetHealthMaterializationDegradedPartitionedMeta(
                        numFailedPartitions=num_failed,
                        numMissingPartitions=num_missing,
                        totalNumPartitions=total_num_partitions,
                    ),
                )
            # missing partitions are ok as long as some partitions are successfully materialized (and no failures)
            # but we want to show the number of missing partitions in the metadata
            return (
                GrapheneAssetHealthStatus.HEALTHY,
                GrapheneAssetHealthMaterializationHealthyPartitionedMeta(
                    numMissingPartitions=num_missing,
                    totalNumPartitions=total_num_partitions,
                )
                if num_missing > 0
                else None,
            )

        asset_record = await AssetRecord.gen(graphene_info.context, asset_key)
        if asset_record is None:
            return GrapheneAssetHealthStatus.UNKNOWN, None
        asset_entry = asset_record.asset_entry

        if self._asset_node_snap.is_observable and not self._asset_node_snap.is_materializable:
            # for observable assets, if there is an observation event then the asset is healthy
            if asset_entry.last_observation is not None:
                return GrapheneAssetHealthStatus.HEALTHY, None
            else:
                return GrapheneAssetHealthStatus.UNKNOWN, None

        if graphene_info.context.instance.can_read_failure_events_for_asset(asset_record):
            # compute the status based on the asset key table
            if (
                asset_entry.last_materialization_storage_id is None
                and asset_entry.last_failed_to_materialize_storage_id is None
            ):
                # never materialized
                return GrapheneAssetHealthStatus.UNKNOWN, None
            if asset_entry.last_failed_to_materialize_storage_id is None:
                # last_materialization_record must be non-null, therefore the asset successfully materialized
                return GrapheneAssetHealthStatus.HEALTHY, None
            elif asset_entry.last_materialization_storage_id is None:
                # last_failed_to_materialize_record must be non-null, therefore the asset failed to materialize
                last_failed_record = check.not_none(asset_entry.last_failed_to_materialize_record)
                return (
                    GrapheneAssetHealthStatus.DEGRADED,
                    GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta(
                        failedRunId=last_failed_record.run_id,
                    ),
                )

            if (
                asset_entry.last_materialization_storage_id
                > asset_entry.last_failed_to_materialize_storage_id
            ):
                # latest materialization succeeded
                return GrapheneAssetHealthStatus.HEALTHY, None
            # latest materialization failed
            last_failed_record = check.not_none(asset_entry.last_failed_to_materialize_record)
            return (
                GrapheneAssetHealthStatus.DEGRADED,
                GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta(
                    failedRunId=last_failed_record.run_id,
                ),
            )
        # we are not storing failure events for this asset, so must compute status based on the information we have available
        # in some cases this results in reporting as asset as HEALTHY or UNKNOWN during an in progress run
        # even if the asset was previously failed
        else:
            # if the asset has been successfully materialized in the past, we fallback to that status
            # when we don't have the information available to compute status based on the latest run
            fallback_status_and_meta = (
                (GrapheneAssetHealthStatus.UNKNOWN, None)
                if asset_entry.last_materialization is None
                else (GrapheneAssetHealthStatus.HEALTHY, None)
            )
            if asset_entry.last_run_id is None:
                return fallback_status_and_meta

            assert asset_entry.last_run_id is not None
            if (
                asset_entry.last_materialization is not None
                and asset_entry.last_run_id == asset_entry.last_materialization.run_id
            ):
                # latest materialization succeeded in the latest run
                return GrapheneAssetHealthStatus.HEALTHY, None
            run_record = await RunRecord.gen(graphene_info.context, asset_entry.last_run_id)
            if run_record is None or not run_record.dagster_run.is_finished:
                return fallback_status_and_meta
            run_end_time = check.not_none(run_record.end_time)
            if (
                asset_entry.last_materialization
                and asset_entry.last_materialization.timestamp > run_end_time
            ):
                # latest materialization was reported manually
                return GrapheneAssetHealthStatus.HEALTHY, None
            if run_record.dagster_run.is_failure:
                return (
                    GrapheneAssetHealthStatus.DEGRADED,
                    GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta(
                        failedRunId=run_record.dagster_run.run_id,
                    ),
                )

            return fallback_status_and_meta

    async def resolve_materializationStatus(self, graphene_info: ResolveInfo):
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                self.get_materialization_status_for_asset_health(graphene_info)
            )
        materialization_status, _ = await self.materialization_status_task
        return materialization_status

    async def resolve_materializationStatusMetadata(self, graphene_info: ResolveInfo):
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                self.get_materialization_status_for_asset_health(graphene_info)
            )
        _, materialization_status_metadata = await self.materialization_status_task
        return materialization_status_metadata

    async def get_asset_check_status_for_asset_health(
        self, graphene_info: ResolveInfo
    ) -> tuple[str, Optional[GrapheneAssetHealthCheckMeta]]:
        """Computes the health indicator for the asset checks for the assets. Follows these rules:
        HEALTHY - the latest completed execution for every check is a success.
        WARNING - the latest completed execution for any asset check failed with severity WARN
            (and no checks failed with severity ERROR).
        DEGRADED - the latest completed execution for any asset check failed with severity ERROR.
        UNKNOWN - any asset checks has never been executed.
        NOT_APPLICABLE - the asset has no asset checks defined.

        Note: the latest completed execution for each check may not have executed based on the
        most recent materialization of the asset.
        """
        remote_check_nodes = graphene_info.context.asset_graph.get_checks_for_asset(
            self._asset_node_snap.asset_key
        )
        if not remote_check_nodes or len(remote_check_nodes) == 0:
            # asset doesn't have checks defined
            return GrapheneAssetHealthStatus.NOT_APPLICABLE, None
        else:
            total_num_checks = len(remote_check_nodes)
            check_statuses = []
            check_failure_severities = []
            asset_check_summary_records = await AssetCheckSummaryRecord.gen_many(
                graphene_info.context,
                [remote_check_node.asset_check.key for remote_check_node in remote_check_nodes],
            )
            for summary_record in asset_check_summary_records:
                if summary_record is None or summary_record.last_check_execution_record is None:
                    # the check has never been executed.
                    continue

                # if the last_check_execution_record is completed, it will be the same as last_completed_check_execution_record,
                # but we check the last_check_execution_record status first since there is an edge case
                # where the record will have status PLANNED, but the resolve_status will be EXECUTION_FAILED
                # because the run for the check failed.
                last_check_execution_status = (
                    await summary_record.last_check_execution_record.resolve_status(
                        graphene_info.context
                    )
                )
                last_check_evaluation = summary_record.last_check_execution_record.evaluation

                if last_check_execution_status in [
                    AssetCheckExecutionResolvedStatus.IN_PROGRESS,
                    AssetCheckExecutionResolvedStatus.SKIPPED,
                ]:
                    # the last check is still in progress or is skipped, so we want to check the status of
                    # the latest completed check instead
                    if summary_record.last_completed_check_execution_record is None:
                        # the check hasn't been executed prior to this in progress check
                        continue
                    last_check_execution_status = (
                        await summary_record.last_completed_check_execution_record.resolve_status(
                            graphene_info.context
                        )
                    )
                    last_check_evaluation = (
                        summary_record.last_completed_check_execution_record.evaluation
                    )

                check_statuses.append(last_check_execution_status)
                if last_check_execution_status == AssetCheckExecutionResolvedStatus.FAILED:
                    # failed checks should always have an evaluation, but default to ERROR if not
                    check_failure_severities.append(
                        last_check_evaluation.severity
                        if last_check_evaluation
                        else AssetCheckSeverity.ERROR
                    )
                if (
                    last_check_execution_status
                    == AssetCheckExecutionResolvedStatus.EXECUTION_FAILED
                ):
                    # EXECUTION_FAILED checks may not have an evaluation, and we want to show these as
                    # degraded health anyway.
                    check_failure_severities.append(AssetCheckSeverity.ERROR)

            num_unexecuted_checks = total_num_checks - len(check_statuses)

            if len(check_statuses) == 0:
                # checks have never been executed
                return GrapheneAssetHealthStatus.UNKNOWN, GrapheneAssetHealthCheckUnknownMeta(
                    numNotExecutedChecks=num_unexecuted_checks,
                    totalNumChecks=total_num_checks,
                )

            if any(
                status == AssetCheckExecutionResolvedStatus.FAILED
                or status == AssetCheckExecutionResolvedStatus.EXECUTION_FAILED
                for status in check_statuses
            ):
                if all(
                    severity == AssetCheckSeverity.WARN for severity in check_failure_severities
                ):
                    # all failed checks are with warning severity
                    return (
                        GrapheneAssetHealthStatus.WARNING,
                        GrapheneAssetHealthCheckWarningMeta(
                            numWarningChecks=len(check_failure_severities),
                            totalNumChecks=total_num_checks,
                        ),
                    )
                # at least one failing check had error severity
                num_failed = len(
                    [sev for sev in check_failure_severities if sev == AssetCheckSeverity.ERROR]
                )
                num_warn = len(check_failure_severities) - num_failed
                return GrapheneAssetHealthStatus.DEGRADED, GrapheneAssetHealthCheckDegradedMeta(
                    numFailedChecks=num_failed,
                    numWarningChecks=num_warn,
                    totalNumChecks=total_num_checks,
                )

            # since we are only looking at the latest completed execution for each check, if there
            # are no failed checks, then all checks must have succeeded
            if num_unexecuted_checks > 0:
                # if any check has never been executed, we report this as unknown, even if other checks
                # have passed
                return (
                    GrapheneAssetHealthStatus.UNKNOWN,
                    GrapheneAssetHealthCheckUnknownMeta(
                        numNotExecutedChecks=num_unexecuted_checks,
                        totalNumChecks=total_num_checks,
                    ),
                )
            # all checks must have executed and passed
            return GrapheneAssetHealthStatus.HEALTHY, None

    async def resolve_assetChecksStatus(self, graphene_info: ResolveInfo):
        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                self.get_asset_check_status_for_asset_health(graphene_info)
            )

        asset_checks_status, _ = await self.asset_check_status_task
        return asset_checks_status

    async def resolve_assetChecksStatusMetadata(self, graphene_info: ResolveInfo):
        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                self.get_asset_check_status_for_asset_health(graphene_info)
            )

        _, asset_checks_status_metadata = await self.asset_check_status_task
        return asset_checks_status_metadata

    async def get_freshness_status_for_asset_health(
        self, graphene_info: ResolveInfo
    ) -> tuple[str, Optional[GrapheneAssetHealthFreshnessMeta]]:
        """Computes the health indicator for the freshness for an asset. Follows these rules:
        HEALTHY - the freshness policy is in a PASS-ing state
        WARNING - the freshness policy is in a WARN-ing state
        DEGRADED - the freshness policy is in a FAIL-ing state
        UNKNOWN - the freshness policy has never been evaluated or is in an UNKNOWN state
        NOT_APPLICABLE - the asset does not have a freshness policy defined.
        """
        if self._asset_node_snap.internal_freshness_policy is None:
            return GrapheneAssetHealthStatus.NOT_APPLICABLE, None

        freshness_state_record = graphene_info.context.instance.get_entity_freshness_state(
            self._asset_node_snap.asset_key
        )
        if freshness_state_record is None:
            return GrapheneAssetHealthStatus.UNKNOWN, None
        state = freshness_state_record.freshness_state
        if state == FreshnessState.PASS:
            return GrapheneAssetHealthStatus.HEALTHY, None

        asset_record = await AssetRecord.gen(graphene_info.context, self._asset_node_snap.asset_key)
        last_materialization = (
            asset_record.asset_entry.last_materialization.timestamp
            if asset_record and asset_record.asset_entry.last_materialization
            else None
        )
        if state == FreshnessState.WARN:
            return GrapheneAssetHealthStatus.WARNING, GrapheneAssetHealthFreshnessMeta(
                lastMaterializedTimestamp=last_materialization,
            )
        if state == FreshnessState.FAIL:
            return GrapheneAssetHealthStatus.DEGRADED, GrapheneAssetHealthFreshnessMeta(
                lastMaterializedTimestamp=last_materialization,
            )

        return GrapheneAssetHealthStatus.UNKNOWN, None

    async def resolve_freshnessStatus(self, graphene_info: ResolveInfo):
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                self.get_freshness_status_for_asset_health(graphene_info)
            )

        freshness_status, _ = await self.freshness_status_task
        return freshness_status

    async def resolve_freshnessStatusMetadata(self, graphene_info: ResolveInfo):
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                self.get_freshness_status_for_asset_health(graphene_info)
            )

        _, freshness_status_metadata = await self.freshness_status_task
        return freshness_status_metadata

    async def resolve_assetHealth(self, graphene_info: ResolveInfo):
        if not graphene_info.context.instance.dagster_observe_supported():
            return GrapheneAssetHealthStatus.UNKNOWN
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                self.get_materialization_status_for_asset_health(graphene_info)
            )
        materialization_status, _ = await self.materialization_status_task
        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                self.get_asset_check_status_for_asset_health(graphene_info)
            )
        asset_checks_status, _ = await self.asset_check_status_task
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                self.get_freshness_status_for_asset_health(graphene_info)
            )
        freshness_status, _ = await self.freshness_status_task
        statuses = [
            materialization_status,
            asset_checks_status,
            freshness_status,
        ]
        if GrapheneAssetHealthStatus.DEGRADED in statuses:
            return GrapheneAssetHealthStatus.DEGRADED
        if GrapheneAssetHealthStatus.WARNING in statuses:
            return GrapheneAssetHealthStatus.WARNING
        # at this point, all statuses are HEALTHY, UNKNOWN, or NOT_APPLICABLE
        if materialization_status == GrapheneAssetHealthStatus.UNKNOWN:
            return GrapheneAssetHealthStatus.UNKNOWN
        if all(
            status == GrapheneAssetHealthStatus.UNKNOWN
            or status == GrapheneAssetHealthStatus.NOT_APPLICABLE
            for status in statuses
        ):
            return GrapheneAssetHealthStatus.UNKNOWN
        # at least one status must be HEALTHY
        return GrapheneAssetHealthStatus.HEALTHY



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import graphene
from dagster._core.definitions.selector import ScheduleSelector

from dagster_graphql.implementation.utils import capture_error
from dagster_graphql.schema.errors import GraphenePythonError, GrapheneScheduleNotFoundError
from dagster_graphql.schema.inputs import GrapheneScheduleSelector
from dagster_graphql.schema.instigation import GrapheneDryRunInstigationTick

if TYPE_CHECKING:
    from dagster_graphql.schema.util import ResolveInfo


class GrapheneScheduleDryRunResult(graphene.Union):
    class Meta:
        types = (
            GrapheneDryRunInstigationTick,
            GraphenePythonError,
            GrapheneScheduleNotFoundError,
        )
        name = "ScheduleDryRunResult"


class GrapheneScheduleDryRunMutation(graphene.Mutation):
    """Enable a schedule to launch runs for a job based on external state change."""

    Output = graphene.NonNull(GrapheneScheduleDryRunResult)

    class Arguments:
        selector_data = graphene.NonNull(GrapheneScheduleSelector)
        timestamp = graphene.Float()

    class Meta:
        name = "ScheduleDryRunMutation"

    @capture_error
    def mutate(
        self, graphene_info: "ResolveInfo", selector_data: Mapping[str, Any], timestamp: float
    ):
        return GrapheneDryRunInstigationTick(
            selector=ScheduleSelector.from_graphql_input(selector_data), timestamp=timestamp
        )


types = [
    GrapheneScheduleDryRunMutation,
    GrapheneScheduleDryRunResult,
]



from typing import Callable

import dagster._check as check
import graphene
from dagster._config.snap import ConfigTypeSnap
from dagster._core.snap import ModeDefSnap

from dagster_graphql.schema.pipelines.logger import GrapheneLogger
from dagster_graphql.schema.pipelines.resource import GrapheneResource
from dagster_graphql.schema.util import ResolveInfo, non_null_list


class GrapheneMode(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    resources = non_null_list(GrapheneResource)
    loggers = non_null_list(GrapheneLogger)

    class Meta:
        name = "Mode"

    def __init__(
        self,
        get_config_type: Callable[[str], ConfigTypeSnap],
        job_graphql_id: str,
        mode_def_snap: ModeDefSnap,
    ):
        super().__init__()
        self._mode_def_snap = check.inst_param(mode_def_snap, "mode_def_snap", ModeDefSnap)
        self._get_config_type = get_config_type
        self._job_graphql_id = job_graphql_id

    def resolve_id(self, _graphene_info: ResolveInfo):
        return f"{self._job_graphql_id}-{self._mode_def_snap.name}"

    def resolve_name(self, _graphene_info: ResolveInfo):
        return self._mode_def_snap.name

    def resolve_description(self, _graphene_info: ResolveInfo):
        return self._mode_def_snap.description

    def resolve_resources(self, _graphene_info: ResolveInfo):
        return [
            GrapheneResource(self._get_config_type, resource_def_snap)
            for resource_def_snap in sorted(self._mode_def_snap.resource_def_snaps)
        ]

    def resolve_loggers(self, _graphene_info: ResolveInfo):
        return [
            GrapheneLogger(self._get_config_type, logger_def_snap)
            for logger_def_snap in sorted(self._mode_def_snap.logger_def_snaps)
        ]



from collections.abc import Sequence
from typing import Optional, cast

import dagster._check as check
from dagster._core.errors import DagsterRunNotFoundError
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation import CodeLocation
from dagster._core.remote_representation.external import RemoteJob
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus
from dagster._core.storage.tags import RESUME_RETRY_TAG
from dagster._core.utils import make_new_run_id
from dagster._core.workspace.context import BaseWorkspaceRequestContext
from dagster._utils.merger import merge_dicts

from dagster_graphql.implementation.external import (
    ensure_valid_config,
    get_remote_execution_plan_or_raise,
)
from dagster_graphql.implementation.utils import ExecutionParams


def _get_run(instance: DagsterInstance, run_id: str) -> DagsterRun:
    run = instance.get_run_by_id(run_id)
    if not run:
        raise DagsterRunNotFoundError(invalid_run_id=run_id)
    return cast("DagsterRun", run)


def compute_step_keys_to_execute(
    graphql_context: BaseWorkspaceRequestContext, execution_params: ExecutionParams
) -> tuple[Optional[Sequence[str]], Optional[KnownExecutionState]]:
    check.inst_param(execution_params, "execution_params", ExecutionParams)

    instance = graphql_context.instance

    if not execution_params.step_keys and is_resume_retry(execution_params):
        # Get step keys from parent_run_id if it's a resume/retry
        parent_run_id = check.not_none(execution_params.execution_metadata.parent_run_id)
        parent_run = _get_run(instance, parent_run_id)
        return KnownExecutionState.build_resume_retry_reexecution(
            instance,
            parent_run,
        )
    else:
        known_state = None
        if execution_params.execution_metadata.parent_run_id and execution_params.step_keys:
            parent_run = _get_run(instance, execution_params.execution_metadata.parent_run_id)
            known_state = KnownExecutionState.build_for_reexecution(
                instance,
                parent_run,
            ).update_for_step_selection(execution_params.step_keys)

        return execution_params.step_keys, known_state


def is_resume_retry(execution_params: ExecutionParams) -> bool:
    check.inst_param(execution_params, "execution_params", ExecutionParams)
    return execution_params.execution_metadata.tags.get(RESUME_RETRY_TAG) == "true"


def create_valid_pipeline_run(
    graphql_context: BaseWorkspaceRequestContext,
    remote_job: RemoteJob,
    execution_params: ExecutionParams,
    code_location: CodeLocation,
) -> DagsterRun:
    ensure_valid_config(remote_job, execution_params.run_config)

    step_keys_to_execute, known_state = compute_step_keys_to_execute(
        graphql_context, execution_params
    )

    execution_plan = get_remote_execution_plan_or_raise(
        graphql_context=graphql_context,
        remote_job=remote_job,
        run_config=execution_params.run_config,
        step_keys_to_execute=step_keys_to_execute,
        known_state=known_state,
    )
    tags = merge_dicts(remote_job.tags, execution_params.execution_metadata.tags)

    dagster_run = graphql_context.instance.create_run(
        job_snapshot=remote_job.job_snapshot,
        execution_plan_snapshot=execution_plan.execution_plan_snapshot,
        parent_job_snapshot=remote_job.parent_job_snapshot,
        job_name=execution_params.selector.job_name,
        run_id=(
            execution_params.execution_metadata.run_id
            if execution_params.execution_metadata.run_id
            else make_new_run_id()
        ),
        asset_selection=(
            frozenset(execution_params.selector.asset_selection)
            if execution_params.selector.asset_selection
            else None
        ),
        asset_check_selection=(
            frozenset(execution_params.selector.asset_check_selection)
            if execution_params.selector.asset_check_selection is not None
            else None
        ),
        op_selection=execution_params.selector.op_selection,
        resolved_op_selection=(
            frozenset(execution_params.selector.op_selection)
            if execution_params.selector.op_selection
            else None
        ),
        run_config=execution_params.run_config,
        step_keys_to_execute=execution_plan.execution_plan_snapshot.step_keys_to_execute,
        tags=tags,
        root_run_id=execution_params.execution_metadata.root_run_id,
        parent_run_id=execution_params.execution_metadata.parent_run_id,
        status=DagsterRunStatus.NOT_STARTED,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
        asset_graph=code_location.get_repository(
            remote_job.repository_handle.repository_name
        ).asset_graph,
    )

    return dagster_run



from collections.abc import Sequence
from typing import Optional

import dagster._check as check
import graphene
from dagster import DefaultSensorStatus
from dagster._core.definitions.selector import SensorSelector
from dagster._core.definitions.sensor_definition import SensorType
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.remote_representation import RemoteSensor, TargetSnap
from dagster._core.remote_representation.external import CompoundID
from dagster._core.scheduler.instigation import InstigatorState, InstigatorStatus
from dagster._core.workspace.permissions import Permissions

from dagster_graphql.implementation.events import iterate_metadata_entries
from dagster_graphql.implementation.fetch_sensors import (
    get_sensor_next_tick,
    reset_sensor,
    set_sensor_cursor,
    start_sensor,
    stop_sensor,
)
from dagster_graphql.implementation.loader import RepositoryScopedBatchLoader
from dagster_graphql.implementation.utils import (
    assert_permission_for_location,
    capture_error,
    require_permission_check,
)
from dagster_graphql.schema.asset_selections import GrapheneAssetSelection
from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.errors import (
    GraphenePythonError,
    GrapheneRepositoryNotFoundError,
    GrapheneSensorNotFoundError,
    GrapheneUnauthorizedError,
)
from dagster_graphql.schema.inputs import GrapheneSensorSelector
from dagster_graphql.schema.instigation import (
    GrapheneDryRunInstigationTick,
    GrapheneInstigationState,
    GrapheneInstigationStatus,
)
from dagster_graphql.schema.metadata import GrapheneMetadataEntry
from dagster_graphql.schema.tags import GrapheneDefinitionTag
from dagster_graphql.schema.util import ResolveInfo, non_null_list


class GrapheneTarget(graphene.ObjectType):
    pipelineName = graphene.NonNull(graphene.String)
    mode = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))

    class Meta:
        name = "Target"

    def __init__(self, target_snap: TargetSnap):
        self._target_snap = check.inst_param(target_snap, "target_snap", TargetSnap)
        super().__init__(
            pipelineName=target_snap.job_name,
            mode=target_snap.mode,
            solidSelection=target_snap.op_selection,
        )


class GrapheneSensorMetadata(graphene.ObjectType):
    assetKeys = graphene.List(graphene.NonNull(GrapheneAssetKey))

    class Meta:
        name = "SensorMetadata"


GrapheneSensorType = graphene.Enum.from_enum(SensorType)


class GrapheneSensor(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    jobOriginId = graphene.NonNull(graphene.String)
    name = graphene.NonNull(graphene.String)
    targets = graphene.List(graphene.NonNull(GrapheneTarget))
    defaultStatus = graphene.NonNull(GrapheneInstigationStatus)
    canReset = graphene.NonNull(graphene.Boolean)
    sensorState = graphene.NonNull(GrapheneInstigationState)
    minIntervalSeconds = graphene.NonNull(graphene.Int)
    description = graphene.String()
    nextTick = graphene.Field(GrapheneDryRunInstigationTick)
    metadata = graphene.NonNull(GrapheneSensorMetadata)
    sensorType = graphene.NonNull(GrapheneSensorType)
    assetSelection = graphene.Field(GrapheneAssetSelection)
    tags = non_null_list(GrapheneDefinitionTag)
    metadataEntries = non_null_list(GrapheneMetadataEntry)

    class Meta:
        name = "Sensor"

    def __init__(
        self,
        remote_sensor: RemoteSensor,
        sensor_state: Optional[InstigatorState],
        batch_loader: Optional[RepositoryScopedBatchLoader] = None,
    ):
        self._remote_sensor = check.inst_param(remote_sensor, "remote_sensor", RemoteSensor)

        # optional run loader, provided by a parent GrapheneRepository object that instantiates
        # multiple sensors
        self._batch_loader = check.opt_inst_param(
            batch_loader, "batch_loader", RepositoryScopedBatchLoader
        )

        self._stored_state = sensor_state
        self._sensor_state = self._remote_sensor.get_current_instigator_state(sensor_state)

        super().__init__(
            name=remote_sensor.name,
            jobOriginId=remote_sensor.get_remote_origin_id(),
            minIntervalSeconds=remote_sensor.min_interval_seconds,
            description=remote_sensor.description,
            targets=[GrapheneTarget(target) for target in remote_sensor.get_targets()],
            metadata=GrapheneSensorMetadata(
                assetKeys=remote_sensor.metadata.asset_keys if remote_sensor.metadata else None
            ),
            sensorType=remote_sensor.sensor_type.value,
            assetSelection=GrapheneAssetSelection(
                asset_selection=remote_sensor.asset_selection,
                repository_handle=remote_sensor.handle.repository_handle,
            )
            if remote_sensor.asset_selection
            else None,
        )

    def resolve_id(self, _) -> str:
        return self._remote_sensor.get_compound_id().to_string()

    def resolve_defaultStatus(self, _graphene_info: ResolveInfo):
        default_sensor_status = self._remote_sensor.default_status

        if default_sensor_status == DefaultSensorStatus.RUNNING:
            return GrapheneInstigationStatus.RUNNING
        elif default_sensor_status == DefaultSensorStatus.STOPPED:
            return GrapheneInstigationStatus.STOPPED

    def resolve_canReset(self, _graphene_info: ResolveInfo):
        return bool(
            self._stored_state and self._stored_state.status != InstigatorStatus.DECLARED_IN_CODE
        )

    def resolve_sensorState(self, _graphene_info: ResolveInfo):
        # forward the batch run loader to the instigation state, which provides the sensor runs
        return GrapheneInstigationState(self._sensor_state, self._batch_loader)

    def resolve_nextTick(self, graphene_info: ResolveInfo):
        return get_sensor_next_tick(graphene_info, self._sensor_state)

    def resolve_tags(self, _graphene_info: ResolveInfo) -> Sequence[GrapheneDefinitionTag]:
        return [
            GrapheneDefinitionTag(key, value)
            for key, value in (self._remote_sensor.tags or {}).items()
        ]

    def resolve_metadataEntries(
        self, _graphene_info: ResolveInfo
    ) -> Sequence[GrapheneMetadataEntry]:
        # Standard metadata is nested under the non-standard ExternalSensorMetadata object for
        # backcompat reasons.
        sensor_metadata = self._remote_sensor.metadata
        if sensor_metadata and sensor_metadata.standard_metadata:
            return list(iterate_metadata_entries(sensor_metadata.standard_metadata))
        else:
            return []


class GrapheneSensorOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneSensor,
            GrapheneSensorNotFoundError,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "SensorOrError"


class GrapheneSensors(graphene.ObjectType):
    results = non_null_list(GrapheneSensor)

    class Meta:
        name = "Sensors"


class GrapheneSensorsOrError(graphene.Union):
    class Meta:
        types = (GrapheneSensors, GrapheneRepositoryNotFoundError, GraphenePythonError)
        name = "SensorsOrError"


class GrapheneStartSensorMutation(graphene.Mutation):
    """Enable a sensor to launch runs for a job based on external state change."""

    Output = graphene.NonNull(GrapheneSensorOrError)

    class Arguments:
        sensor_selector = graphene.NonNull(GrapheneSensorSelector)

    class Meta:
        name = "StartSensorMutation"

    @capture_error
    @require_permission_check(Permissions.UPDATE_SENSOR_CURSOR)
    def mutate(self, graphene_info: ResolveInfo, sensor_selector):
        selector = SensorSelector.from_graphql_input(sensor_selector)
        assert_permission_for_location(
            graphene_info, Permissions.UPDATE_SENSOR_CURSOR, selector.location_name
        )
        return start_sensor(graphene_info, selector)


class GrapheneStopSensorMutationResult(graphene.ObjectType):
    instigationState = graphene.Field(GrapheneInstigationState)

    class Meta:
        name = "StopSensorMutationResult"

    def __init__(self, instigator_state):
        super().__init__()
        self._instigator_state = check.inst_param(
            instigator_state, "instigator_state", InstigatorState
        )

    def resolve_instigationState(self, _graphene_info: ResolveInfo):
        return GrapheneInstigationState(instigator_state=self._instigator_state)


class GrapheneStopSensorMutationResultOrError(graphene.Union):
    class Meta:
        types = (GrapheneStopSensorMutationResult, GrapheneUnauthorizedError, GraphenePythonError)
        name = "StopSensorMutationResultOrError"


class GrapheneStopSensorMutation(graphene.Mutation):
    """Disable a sensor from launching runs for a job."""

    Output = graphene.NonNull(GrapheneStopSensorMutationResultOrError)

    class Arguments:
        id = graphene.Argument(graphene.String)  # Sensor / InstigationState id

        # "job" legacy name for instigators, predates current Job
        job_origin_id = graphene.Argument(graphene.String)
        job_selector_id = graphene.Argument(graphene.String)

    class Meta:
        name = "StopSensorMutation"

    @capture_error
    @require_permission_check(Permissions.EDIT_SENSOR)
    def mutate(
        self,
        graphene_info: ResolveInfo,
        id: Optional[str] = None,
        job_origin_id: Optional[str] = None,
        job_selector_id: Optional[str] = None,
    ):
        if id:
            cid = CompoundID.from_string(id)
            sensor_origin_id = cid.remote_origin_id
            sensor_selector_id = cid.selector_id
        elif job_origin_id and CompoundID.is_valid_string(job_origin_id):
            # cross-push handle if InstigationState.id being passed through as origin id
            cid = CompoundID.from_string(job_origin_id)
            sensor_origin_id = cid.remote_origin_id
            sensor_selector_id = cid.selector_id
        elif job_origin_id is None or job_selector_id is None:
            raise DagsterInvariantViolationError("Must specify id or jobOriginId and jobSelectorId")
        else:
            sensor_origin_id = job_origin_id
            sensor_selector_id = job_selector_id

        return stop_sensor(graphene_info, sensor_origin_id, sensor_selector_id)


class GrapheneResetSensorMutation(graphene.Mutation):
    """Reset a sensor to its status defined in code, otherwise disable it from launching runs for a job."""

    Output = graphene.NonNull(GrapheneSensorOrError)

    class Arguments:
        sensor_selector = graphene.NonNull(GrapheneSensorSelector)

    class Meta:
        name = "ResetSensorMutation"

    @capture_error
    @require_permission_check(Permissions.EDIT_SENSOR)
    def mutate(self, graphene_info: ResolveInfo, sensor_selector):
        selector = SensorSelector.from_graphql_input(sensor_selector)

        assert_permission_for_location(
            graphene_info, Permissions.EDIT_SENSOR, selector.location_name
        )

        return reset_sensor(graphene_info, selector)


class GrapheneSetSensorCursorMutation(graphene.Mutation):
    """Set a cursor for a sensor to track state across evaluations."""

    Output = graphene.NonNull(GrapheneSensorOrError)

    class Arguments:
        sensor_selector = graphene.NonNull(GrapheneSensorSelector)
        cursor = graphene.String()

    class Meta:
        name = "SetSensorCursorMutation"

    @capture_error
    def mutate(self, graphene_info: ResolveInfo, sensor_selector, cursor=None):
        selector = SensorSelector.from_graphql_input(sensor_selector)
        assert_permission_for_location(
            graphene_info, Permissions.UPDATE_SENSOR_CURSOR, selector.location_name
        )
        return set_sensor_cursor(graphene_info, selector, cursor)


types = [
    GrapheneSensor,
    GrapheneSensorOrError,
    GrapheneSensors,
    GrapheneSensorsOrError,
    GrapheneStopSensorMutation,
    GrapheneStopSensorMutationResult,
    GrapheneStopSensorMutationResultOrError,
    GrapheneStopSensorMutation,
    GrapheneSetSensorCursorMutation,
    GrapheneResetSensorMutation,
]



import json
import logging
from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional

import dagster._check as check
import graphene
from dagster import AssetKey
from dagster._core.definitions.backfill_policy import BackfillPolicy, BackfillPolicyType
from dagster._core.definitions.partition import PartitionsSubset
from dagster._core.definitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsSubset
from dagster._core.errors import DagsterError, DagsterInvariantViolationError
from dagster._core.execution.asset_backfill import (
    AssetBackfillStatus,
    PartitionedAssetBackfillStatus,
    UnpartitionedAssetBackfillStatus,
)
from dagster._core.execution.backfill import BulkActionStatus, PartitionBackfill
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation.external import RemotePartitionSet
from dagster._core.storage.compute_log_manager import ComputeIOType
from dagster._core.storage.dagster_run import DagsterRun, RunPartitionData, RunRecord, RunsFilter
from dagster._core.storage.tags import (
    ASSET_PARTITION_RANGE_END_TAG,
    ASSET_PARTITION_RANGE_START_TAG,
    TagType,
    get_tag_type,
)
from dagster._core.workspace.permissions import Permissions
from dagster_shared import seven

from dagster_graphql.implementation.fetch_partition_sets import (
    partition_status_counts_from_run_partition_data,
    partition_statuses_from_run_partition_data,
)
from dagster_graphql.implementation.utils import has_permission_for_asset_graph
from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.errors import (
    GrapheneError,
    GrapheneInvalidOutputError,
    GrapheneInvalidStepError,
    GrapheneInvalidSubsetError,
    GraphenePartitionKeysNotFoundError,
    GraphenePartitionSetNotFoundError,
    GraphenePipelineNotFoundError,
    GraphenePythonError,
    GrapheneRunConflict,
    GrapheneUnauthorizedError,
    create_execution_params_error_types,
)
from dagster_graphql.schema.pipelines.config import GrapheneRunConfigValidationInvalid
from dagster_graphql.schema.pipelines.status import GrapheneRunStatus
from dagster_graphql.schema.runs_feed import GrapheneRunsFeedEntry
from dagster_graphql.schema.util import ResolveInfo, non_null_list

if TYPE_CHECKING:
    from dagster_graphql.schema.partition_sets import (
        GraphenePartitionSet,
        GraphenePartitionStatusCounts,
    )
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

pipeline_execution_error_types = (
    GrapheneInvalidStepError,
    GrapheneInvalidOutputError,
    GrapheneRunConfigValidationInvalid,
    GraphenePipelineNotFoundError,
    GrapheneRunConflict,
    GrapheneUnauthorizedError,
    GraphenePythonError,
    GrapheneInvalidSubsetError,
) + create_execution_params_error_types


class GrapheneLaunchBackfillSuccess(graphene.ObjectType):
    backfill_id = graphene.NonNull(graphene.String)
    launched_run_ids = graphene.List(graphene.String)

    class Meta:
        name = "LaunchBackfillSuccess"


class GrapheneLaunchBackfillResult(graphene.Union):
    class Meta:
        types = (
            GrapheneLaunchBackfillSuccess,
            GraphenePartitionSetNotFoundError,
            GraphenePartitionKeysNotFoundError,
        ) + pipeline_execution_error_types
        name = "LaunchBackfillResult"


class GrapheneCancelBackfillSuccess(graphene.ObjectType):
    backfill_id = graphene.NonNull(graphene.String)

    class Meta:
        name = "CancelBackfillSuccess"


class GrapheneCancelBackfillResult(graphene.Union):
    class Meta:
        types = (GrapheneCancelBackfillSuccess, GrapheneUnauthorizedError, GraphenePythonError)
        name = "CancelBackfillResult"


class GrapheneResumeBackfillSuccess(graphene.ObjectType):
    backfill_id = graphene.NonNull(graphene.String)

    class Meta:
        name = "ResumeBackfillSuccess"


class GrapheneResumeBackfillResult(graphene.Union):
    class Meta:
        types = (GrapheneResumeBackfillSuccess, GrapheneUnauthorizedError, GraphenePythonError)
        name = "ResumeBackfillResult"


class GrapheneBulkActionStatus(graphene.Enum):
    REQUESTED = "REQUESTED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELED = "CANCELED"
    CANCELING = "CANCELING"
    COMPLETED_SUCCESS = "COMPLETED_SUCCESS"
    COMPLETED_FAILED = "COMPLETED_FAILED"

    class Meta:
        name = "BulkActionStatus"

    def to_dagster_run_status(self) -> GrapheneRunStatus:
        """Maps bulk action status to a run status for use with the RunsFeedEntry interface."""
        # the pyright ignores are required because GrapheneBulkActionStatus.STATUS and GrapheneRunStatus.STATUS
        # are interpreted as a Literal string during static analysis, but it is actually an Enum value
        if self.args[0] == GrapheneBulkActionStatus.REQUESTED.value:  # pyright: ignore[reportAttributeAccessIssue]
            return GrapheneRunStatus.STARTED  # pyright: ignore[reportReturnType]
        if self.args[0] == GrapheneBulkActionStatus.COMPLETED.value:  # pyright: ignore[reportAttributeAccessIssue]
            return GrapheneRunStatus.SUCCESS  # pyright: ignore[reportReturnType]
        if self.args[0] == GrapheneBulkActionStatus.COMPLETED_SUCCESS.value:  # pyright: ignore[reportAttributeAccessIssue]
            return GrapheneRunStatus.SUCCESS  # pyright: ignore[reportReturnType]
        if self.args[0] == GrapheneBulkActionStatus.COMPLETED_FAILED.value:  # pyright: ignore[reportAttributeAccessIssue]
            return GrapheneRunStatus.FAILURE  # pyright: ignore[reportReturnType]
        if self.args[0] == GrapheneBulkActionStatus.FAILED.value:  # pyright: ignore[reportAttributeAccessIssue]
            return GrapheneRunStatus.FAILURE  # pyright: ignore[reportReturnType]
        if self.args[0] == GrapheneBulkActionStatus.CANCELED.value:  # pyright: ignore[reportAttributeAccessIssue]
            return GrapheneRunStatus.CANCELED  # pyright: ignore[reportReturnType]
        if self.args[0] == GrapheneBulkActionStatus.CANCELING.value:  # pyright: ignore[reportAttributeAccessIssue]
            return GrapheneRunStatus.CANCELING  # pyright: ignore[reportReturnType]

        raise DagsterInvariantViolationError(
            f"Unable to convert BulkActionStatus {self.args[0]} to a RunStatus. {self.args[0]} is an unknown status."
        )


class GrapheneAssetBackfillTargetPartitions(graphene.ObjectType):
    class Meta:
        name = "AssetBackfillTargetPartitions"

    ranges = graphene.List(
        graphene.NonNull("dagster_graphql.schema.partition_sets.GraphenePartitionKeyRange")
    )
    partitionKeys = graphene.List(graphene.NonNull(graphene.String))

    def __init__(self, partition_subset: PartitionsSubset):
        from dagster_graphql.schema.partition_sets import GraphenePartitionKeyRange

        if isinstance(partition_subset, TimeWindowPartitionsSubset):
            ranges = [
                GraphenePartitionKeyRange(start, end)
                for start, end in partition_subset.get_partition_key_ranges(
                    partition_subset.partitions_def
                )
            ]
            partition_keys = None
        else:  # Default partitions subset
            ranges = None
            partition_keys = partition_subset.get_partition_keys()

        super().__init__(
            ranges=ranges,
            partitionKeys=partition_keys,
        )


class GrapheneAssetPartitions(graphene.ObjectType):
    assetKey = graphene.NonNull(GrapheneAssetKey)
    partitions = graphene.Field(GrapheneAssetBackfillTargetPartitions)

    class Meta:
        name = "AssetPartitions"

    def __init__(self, asset_key: AssetKey, partitions_subset: Optional[PartitionsSubset]):
        if partitions_subset is None:
            partitions = None
        else:
            partitions = GrapheneAssetBackfillTargetPartitions(partitions_subset)

        super().__init__(assetKey=GrapheneAssetKey(path=asset_key.path), partitions=partitions)


class GraphenePartitionRange(graphene.ObjectType):
    start = graphene.NonNull(graphene.String)
    end = graphene.NonNull(graphene.String)

    class Meta:
        name = "PartitionRange"

    def __init__(self, start: str, end: str):
        super().__init__(start=start, end=end)


class GrapheneAssetPartitionRange(graphene.ObjectType):
    assetKey = graphene.NonNull(GrapheneAssetKey)
    partitionRange = graphene.Field(GraphenePartitionRange)

    class Meta:
        name = "AssetPartitionRange"

    def __init__(self, asset_key: AssetKey, partition_range: Optional[PartitionKeyRange]):
        super().__init__(
            assetKey=GrapheneAssetKey(path=asset_key.path),
            partitionRange=(
                GraphenePartitionRange(start=partition_range.start, end=partition_range.end)
                if partition_range
                else None
            ),
        )


class GrapheneAssetBackfillData(graphene.ObjectType):
    class Meta:
        name = "AssetBackfillData"

    def __init__(self, backfill_job: PartitionBackfill):
        self._backfill_job = backfill_job
        check.invariant(self._backfill_job.is_asset_backfill, "Must be an asset backfill")

    assetBackfillStatuses = non_null_list(
        "dagster_graphql.schema.partition_sets.GrapheneAssetBackfillStatus"
    )
    rootTargetedPartitions = graphene.Field(
        "dagster_graphql.schema.backfill.GrapheneAssetBackfillTargetPartitions",
    )

    def resolve_rootTargetedPartitions(self, graphene_info: ResolveInfo):
        try:
            root_partitions_subset = self._backfill_job.get_target_root_partitions_subset(
                graphene_info.context
            )
        except DagsterError:
            logging.getLogger("dagster").warning(
                "Error generating root target partitions", exc_info=True
            )
            return None

        return (
            GrapheneAssetBackfillTargetPartitions(root_partitions_subset)
            if root_partitions_subset
            else None
        )

    def resolve_assetBackfillStatuses(self, graphene_info: ResolveInfo):
        from dagster_graphql.schema.partition_sets import (
            GrapheneAssetPartitionsStatusCounts,
            GrapheneUnpartitionedAssetStatus,
        )

        status_per_asset = self._backfill_job.get_backfill_status_per_asset_key(
            graphene_info.context
        )

        asset_partition_status_counts = []

        for asset_status in status_per_asset:
            if isinstance(asset_status, PartitionedAssetBackfillStatus):
                asset_partition_status_counts.append(
                    GrapheneAssetPartitionsStatusCounts(
                        assetKey=asset_status.asset_key,
                        numPartitionsTargeted=asset_status.num_targeted_partitions,
                        numPartitionsInProgress=asset_status.partitions_counts_by_status[
                            AssetBackfillStatus.IN_PROGRESS
                        ],
                        numPartitionsMaterialized=asset_status.partitions_counts_by_status[
                            AssetBackfillStatus.MATERIALIZED
                        ],
                        numPartitionsFailed=asset_status.partitions_counts_by_status[
                            AssetBackfillStatus.FAILED
                        ],
                    )
                )
            else:
                if not isinstance(asset_status, UnpartitionedAssetBackfillStatus):
                    check.failed(f"Unexpected asset status type {type(asset_status)}")

                asset_partition_status_counts.append(
                    GrapheneUnpartitionedAssetStatus(
                        assetKey=asset_status.asset_key,
                        inProgress=asset_status.backfill_status is AssetBackfillStatus.IN_PROGRESS,
                        materialized=asset_status.backfill_status
                        is AssetBackfillStatus.MATERIALIZED,
                        failed=asset_status.backfill_status is AssetBackfillStatus.FAILED,
                    )
                )

        return asset_partition_status_counts


class GraphenePartitionBackfill(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneRunsFeedEntry,)

        name = "PartitionBackfill"

    id = graphene.NonNull(graphene.ID)
    status = graphene.NonNull(GrapheneBulkActionStatus)
    runStatus = graphene.Field(
        graphene.NonNull(
            GrapheneRunStatus,
            description="Included to comply with RunsFeedEntry interface. Synthesizes status and the status of subruns to create a RunStatus.",
        )
    )
    partitionNames = graphene.List(graphene.NonNull(graphene.String))
    isValidSerialization = graphene.NonNull(graphene.Boolean)
    numPartitions = graphene.Field(graphene.Int)
    numCancelable = graphene.NonNull(graphene.Int)
    fromFailure = graphene.NonNull(graphene.Boolean)
    reexecutionSteps = graphene.List(graphene.NonNull(graphene.String))
    assetSelection = graphene.List(graphene.NonNull(GrapheneAssetKey))
    partitionSetName = graphene.Field(graphene.String)
    timestamp = graphene.NonNull(graphene.Float)
    creationTime = graphene.Field(
        graphene.NonNull(
            graphene.Float,
            description="Included to comply with RunsFeedEntry interface. Duplicate of timestamp.",
        )
    )
    startTime = graphene.Field(
        graphene.Float(),
        description="Included to comply with RunsFeedEntry interface. Duplicate of timestamp.",
    )
    endTimestamp = graphene.Field(graphene.Float)
    endTime = graphene.Field(
        graphene.Float(),
        description="Included to comply with RunsFeedEntry interface. Duplicate of endTimestamp.",
    )
    partitionSet = graphene.Field("dagster_graphql.schema.partition_sets.GraphenePartitionSet")
    runs = graphene.Field(
        non_null_list("dagster_graphql.schema.pipelines.pipeline.GrapheneRun"),
        limit=graphene.Int(),
    )
    unfinishedRuns = graphene.Field(
        non_null_list("dagster_graphql.schema.pipelines.pipeline.GrapheneRun"),
        limit=graphene.Int(),
    )
    cancelableRuns = graphene.Field(
        non_null_list("dagster_graphql.schema.pipelines.pipeline.GrapheneRun"), limit=graphene.Int()
    )
    error = graphene.Field(GraphenePythonError)
    partitionStatuses = graphene.Field(
        "dagster_graphql.schema.partition_sets.GraphenePartitionStatuses"
    )
    partitionStatusCounts = non_null_list(
        "dagster_graphql.schema.partition_sets.GraphenePartitionStatusCounts"
    )
    partitionsTargetedForAssetKey = graphene.Field(
        "dagster_graphql.schema.backfill.GrapheneAssetBackfillTargetPartitions",
        asset_key=graphene.Argument("dagster_graphql.schema.inputs.GrapheneAssetKeyInput"),
    )
    isAssetBackfill = graphene.NonNull(graphene.Boolean)
    assetBackfillData = graphene.Field(GrapheneAssetBackfillData)

    hasCancelPermission = graphene.NonNull(graphene.Boolean)
    hasResumePermission = graphene.NonNull(graphene.Boolean)
    user = graphene.Field(graphene.String)
    tags = non_null_list("dagster_graphql.schema.tags.GraphenePipelineTag")
    title = graphene.Field(graphene.String)
    description = graphene.Field(graphene.String)
    logEvents = graphene.Field(
        graphene.NonNull("dagster_graphql.schema.instigation.GrapheneInstigationEventConnection"),
        cursor=graphene.String(),
    )
    jobName = graphene.Field(
        graphene.String(),
        description="Included to comply with RunsFeedEntry interface.",
    )
    assetCheckSelection = graphene.List(
        graphene.NonNull("dagster_graphql.schema.entity_key.GrapheneAssetCheckHandle")
    )

    def __init__(self, backfill_job: PartitionBackfill):
        self._backfill_job = check.inst_param(backfill_job, "backfill_job", PartitionBackfill)

        self._records = None
        self._partition_run_data = None

        super().__init__(
            id=backfill_job.backfill_id,
            partitionSetName=backfill_job.partition_set_name,
            jobName=backfill_job.job_name,
            status=backfill_job.status.value,
            fromFailure=bool(backfill_job.from_failure),
            reexecutionSteps=backfill_job.reexecution_steps,
            timestamp=backfill_job.backfill_timestamp,
            startTime=backfill_job.backfill_timestamp,
            creationTime=backfill_job.backfill_timestamp,
            assetSelection=backfill_job.asset_selection,
            assetCheckSelection=[],
        )

    def _get_partition_set(self, graphene_info: ResolveInfo) -> Optional[RemotePartitionSet]:
        if self._backfill_job.partition_set_origin is None:
            return None

        origin = self._backfill_job.partition_set_origin
        location_name = origin.repository_origin.code_location_origin.location_name
        repository_name = origin.repository_origin.repository_name
        if not graphene_info.context.has_code_location(location_name):
            return None

        location = graphene_info.context.get_code_location(location_name)
        if not location.has_repository(repository_name):
            return None

        repository = location.get_repository(repository_name)
        partition_sets = [
            partition_set
            for partition_set in repository.get_partition_sets()
            if partition_set.name == origin.partition_set_name
        ]
        if not partition_sets:
            return None

        return partition_sets[0]

    def _get_records(self, graphene_info: ResolveInfo) -> Sequence[RunRecord]:
        if self._records is None:
            filters = RunsFilter.for_backfill(self._backfill_job.backfill_id)
            self._records = graphene_info.context.instance.get_run_records(
                filters=filters,
            )
        return self._records

    def _get_partition_run_data(self, graphene_info: ResolveInfo) -> Sequence[RunPartitionData]:
        if self._partition_run_data is not None:
            return self._partition_run_data

        # Storage layer not equipped to calculate partition status for ranged job backfills so
        # we perform the calculation ad hoc here.
        partition_set = self._get_partition_set(graphene_info)
        if (
            partition_set
            and partition_set.backfill_policy
            and partition_set.backfill_policy.max_partitions_per_run != 1
        ):  # ranged backfills
            self._partition_run_data = self._get_partition_run_data_for_ranged_job_backfill(
                graphene_info.context.instance, partition_set
            )
        else:
            self._partition_run_data = (
                graphene_info.context.instance.run_storage.get_run_partition_data(
                    runs_filter=RunsFilter(
                        tags=DagsterRun.tags_for_backfill_id(self._backfill_job.backfill_id)
                    )
                )
            )
        return self._partition_run_data

    def _get_partition_run_data_for_ranged_job_backfill(
        self, instance: DagsterInstance, partition_set: RemotePartitionSet
    ) -> Sequence[RunPartitionData]:
        partitions_def = partition_set.get_partitions_definition()
        records = instance.get_run_records(
            RunsFilter(tags=DagsterRun.tags_for_backfill_id(self._backfill_job.backfill_id))
        )
        return [
            RunPartitionData(
                run_id=record.dagster_run.run_id,
                partition=key,
                status=record.dagster_run.status,
                start_time=record.start_time,
                end_time=record.end_time,
            )
            for record in records
            for key in partitions_def.get_partition_keys_in_range(
                PartitionKeyRange(
                    start=record.dagster_run.tags[ASSET_PARTITION_RANGE_START_TAG],
                    end=record.dagster_run.tags[ASSET_PARTITION_RANGE_END_TAG],
                ),
                instance,
            )
        ]

    @property
    def creation_timestamp(self) -> float:
        return self.timestamp

    def resolve_unfinishedRuns(self, graphene_info: ResolveInfo) -> Sequence["GrapheneRun"]:
        from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

        records = self._get_records(graphene_info)
        return [GrapheneRun(record) for record in records if not record.dagster_run.is_finished]

    def resolve_cancelableRuns(self, graphene_info: ResolveInfo) -> Sequence["GrapheneRun"]:
        from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

        records = self._get_records(graphene_info)
        return [GrapheneRun(record) for record in records if record.dagster_run.is_cancelable]

    def resolve_runs(self, graphene_info: ResolveInfo) -> "Sequence[GrapheneRun]":
        from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

        records = self._get_records(graphene_info)
        return [GrapheneRun(record) for record in records]

    def resolve_tags(self, _graphene_info: ResolveInfo):
        from dagster_graphql.schema.tags import GraphenePipelineTag

        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in self._backfill_job.tags.items()
            if get_tag_type(key) != TagType.HIDDEN
        ]

    def resolve_runStatus(self, _graphene_info: ResolveInfo) -> GrapheneRunStatus:
        return GrapheneBulkActionStatus(self.status).to_dagster_run_status()

    def resolve_endTimestamp(self, graphene_info: ResolveInfo) -> Optional[float]:
        if self._backfill_job.backfill_end_timestamp is not None:
            return self._backfill_job.backfill_end_timestamp
        if self._backfill_job.status == BulkActionStatus.REQUESTED:
            # if it's still in progress then there is no end time
            return None
        records = self._get_records(graphene_info)
        if len(records) == 0:
            # backfill was moved to a terminal state before any runs were launched. We cannot
            # reconstruct the time the backfill actually moved to a terminal state, so use the start
            # time as an estimation
            return self.creationTime
        max_end_time = 0
        for record in records:
            max_end_time = max(record.end_time or 0, max_end_time)

        return max_end_time

    def resolve_endTime(self, graphene_info: ResolveInfo) -> Optional[float]:
        return self.resolve_endTimestamp(graphene_info)

    def resolve_isValidSerialization(self, _graphene_info: ResolveInfo) -> bool:
        return self._backfill_job.is_valid_serialization(_graphene_info.context)

    def resolve_partitionNames(self, _graphene_info: ResolveInfo) -> Optional[Sequence[str]]:
        return self._backfill_job.get_partition_names(_graphene_info.context)

    def resolve_numPartitions(self, _graphene_info: ResolveInfo) -> Optional[int]:
        return self._backfill_job.get_num_partitions(_graphene_info.context)

    def resolve_numCancelable(self, _graphene_info: ResolveInfo) -> int:
        return self._backfill_job.get_num_cancelable()

    def resolve_partitionSet(self, graphene_info: ResolveInfo) -> Optional["GraphenePartitionSet"]:
        from dagster_graphql.schema.partition_sets import GraphenePartitionSet

        partition_set = self._get_partition_set(graphene_info)

        if not partition_set:
            return None

        return GraphenePartitionSet(
            repository_handle=partition_set.repository_handle,
            remote_partition_set=partition_set,
        )

    def resolve_partitionStatuses(self, graphene_info: ResolveInfo):
        if self._backfill_job.is_asset_backfill:
            return None

        partition_set_origin = self._backfill_job.partition_set_origin
        partition_set_name = (
            partition_set_origin.partition_set_name if partition_set_origin else None
        )
        partition_run_data = self._get_partition_run_data(graphene_info)
        return partition_statuses_from_run_partition_data(
            partition_set_name,
            partition_run_data,
            check.not_none(self._backfill_job.get_partition_names(graphene_info.context)),
            backfill_id=self._backfill_job.backfill_id,
        )

    def resolve_partitionStatusCounts(
        self, graphene_info: ResolveInfo
    ) -> Sequence["GraphenePartitionStatusCounts"]:
        # This resolver is only enabled for job backfills, since it assumes a unique run per
        # partition key (which is not true for asset backfills). Asset backfills should rely on
        # the assetBackfillData resolver instead.
        if self._backfill_job.is_asset_backfill:
            return []
        else:
            partition_run_data = self._get_partition_run_data(graphene_info)
            return partition_status_counts_from_run_partition_data(
                partition_run_data,
                check.not_none(self._backfill_job.get_partition_names(graphene_info.context)),
            )

    def resolve_isAssetBackfill(self, _graphene_info: ResolveInfo) -> bool:
        return self._backfill_job.is_asset_backfill

    def resolve_partitionsTargetedForAssetKey(
        self, graphene_info: ResolveInfo, asset_key
    ) -> Optional[PartitionsSubset]:
        from dagster._core.definitions.events import AssetKey

        if not self._backfill_job.is_asset_backfill:
            return None

        root_partitions_subset = self._backfill_job.get_target_partitions_subset(
            graphene_info.context, AssetKey.from_graphql_input(asset_key)
        )
        if not root_partitions_subset:
            return None

        return GrapheneAssetBackfillTargetPartitions(root_partitions_subset)

    def resolve_assetBackfillData(
        self, graphene_info: ResolveInfo
    ) -> Optional[GrapheneAssetBackfillData]:
        if not self._backfill_job.is_asset_backfill:
            return None

        return GrapheneAssetBackfillData(self._backfill_job)

    def resolve_error(self, _graphene_info: ResolveInfo) -> Optional[GraphenePythonError]:
        if self._backfill_job.error:
            return GraphenePythonError(self._backfill_job.error)
        return None

    def resolve_hasCancelPermission(self, graphene_info: ResolveInfo) -> bool:
        if self._backfill_job.is_asset_backfill:
            return has_permission_for_asset_graph(
                graphene_info,
                graphene_info.context.asset_graph,
                self._backfill_job.asset_selection,
                Permissions.CANCEL_PARTITION_BACKFILL,
            )
        if self._backfill_job.partition_set_origin is None:
            return graphene_info.context.has_permission(Permissions.CANCEL_PARTITION_BACKFILL)
        location_name = self._backfill_job.partition_set_origin.selector.location_name
        return graphene_info.context.has_permission_for_location(
            Permissions.CANCEL_PARTITION_BACKFILL, location_name
        )

    def resolve_hasResumePermission(self, graphene_info: ResolveInfo) -> bool:
        if self._backfill_job.is_asset_backfill:
            return has_permission_for_asset_graph(
                graphene_info,
                graphene_info.context.asset_graph,
                self._backfill_job.asset_selection,
                Permissions.LAUNCH_PARTITION_BACKFILL,
            )

        if self._backfill_job.partition_set_origin is None:
            return graphene_info.context.has_permission(Permissions.LAUNCH_PARTITION_BACKFILL)
        location_name = self._backfill_job.partition_set_origin.selector.location_name
        return graphene_info.context.has_permission_for_location(
            Permissions.LAUNCH_PARTITION_BACKFILL, location_name
        )

    def resolve_user(self, _graphene_info: ResolveInfo) -> Optional[str]:
        return self._backfill_job.user

    def resolve_title(self, _graphene_info: ResolveInfo) -> Optional[str]:
        return self._backfill_job.title

    def resolve_description(self, _graphene_info: ResolveInfo) -> Optional[str]:
        return self._backfill_job.description

    def resolve_logEvents(self, graphene_info: ResolveInfo, cursor: Optional[str] = None):
        from dagster_graphql.schema.instigation import (
            GrapheneInstigationEvent,
            GrapheneInstigationEventConnection,
        )
        from dagster_graphql.schema.logs.log_level import GrapheneLogLevel

        backfill_log_key_prefix = self._backfill_job.log_storage_prefix

        instance = graphene_info.context.instance

        if not instance.backfill_log_storage_enabled():
            return GrapheneInstigationEventConnection(events=[], cursor="", hasMore=False)

        records, new_cursor = instance.compute_log_manager.read_log_lines_for_log_key_prefix(
            backfill_log_key_prefix, cursor, io_type=ComputeIOType.STDERR
        )

        events = []
        for line in records:
            if not line:
                continue
            try:
                record_dict = seven.json.loads(line)
            except json.JSONDecodeError:
                continue

            exc_info = record_dict.get("exc_info")
            message = record_dict.get("msg")
            if exc_info:
                message = f"{message}\n\n{exc_info}"
            event = GrapheneInstigationEvent(
                message=message,
                level=GrapheneLogLevel.from_level(record_dict["levelno"]),
                timestamp=int(record_dict["created"] * 1000),
            )

            events.append(event)

        return GrapheneInstigationEventConnection(
            events=events,
            cursor=new_cursor.to_string() if new_cursor else "",
            hasMore=new_cursor.has_more_now if new_cursor else False,
        )


class GrapheneBackfillNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "BackfillNotFoundError"

    backfill_id = graphene.NonNull(graphene.String)

    def __init__(self, backfill_id: str):
        super().__init__()
        self.backfill_id = backfill_id
        self.message = f"Backfill {backfill_id} could not be found."


class GraphenePartitionBackfillOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitionBackfill, GrapheneBackfillNotFoundError, GraphenePythonError)
        name = "PartitionBackfillOrError"


class GraphenePartitionBackfills(graphene.ObjectType):
    results = non_null_list(GraphenePartitionBackfill)

    class Meta:
        name = "PartitionBackfills"


class GraphenePartitionBackfillsOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitionBackfills, GraphenePythonError)
        name = "PartitionBackfillsOrError"


GrapheneBackfillPolicyType = graphene.Enum.from_enum(BackfillPolicyType)


class GrapheneBackfillPolicy(graphene.ObjectType):
    maxPartitionsPerRun = graphene.Field(graphene.Int())
    description = graphene.NonNull(graphene.String)
    policyType = graphene.NonNull(GrapheneBackfillPolicyType)

    class Meta:
        name = "BackfillPolicy"

    def __init__(self, backfill_policy: BackfillPolicy):
        self._backfill_policy = check.inst_param(backfill_policy, "backfill_policy", BackfillPolicy)
        super().__init__(
            maxPartitionsPerRun=backfill_policy.max_partitions_per_run,
            policyType=backfill_policy.policy_type,
        )

    def resolve_description(self, _graphene_info: ResolveInfo) -> str:
        if self._backfill_policy.max_partitions_per_run is None:
            return "Backfills all partitions in a single run"
        else:
            return f"Backfills in multiple runs, with a maximum of {self._backfill_policy.max_partitions_per_run} partitions per run"
